// CourseLoader.cpp: implementation of the CCourseLoader class.
//
//////////////////////////////////////////////////////////////////////

#include "CourseLoader.h"
#include "GameMath.h"
#include "GoalFactory.h"
#include <DXErr8.h>
#include <crtdbg.h>     //for _ASSERT
#include <fstream>
using namespace std;

VECTOR2TEX g_pWallSect[] =
{
    VECTOR2TEX(-0.1f,  0.0f,   0.00f),
    VECTOR2TEX(-0.1f,  10.0f,  0.75f),
    VECTOR2TEX(0.1f,  10.0f,  0.25f),
    VECTOR2TEX(0.1f,  0.0f,   0.50f)
};

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CCourseLoader::CCourseLoader(const char* pszAppRegKey) : 
                                 m_pd3dDevice(NULL),
                                 m_iWallsSize(0),
                                 m_pWalls(NULL),
                                 m_iMeshObjectSize(0),
                                 m_pMeshObject(NULL),
                                 m_iMeshPlaceSize(0),
                                 m_pMeshPlace(NULL),
                                 m_pTerrain(NULL),
                                 m_pCourseMesh(NULL),
                                 m_pSkyBox(NULL),
                                 m_iHighScore(0),
                                 m_pWaterPlane(NULL),
                                 m_pSnakeStarts(NULL),
                                 m_pSnakeMoveDirs(NULL),
                                 m_iSnakeStarts(0)
{
    m_pszHighName[0] = '\0';
    m_pszFilename[0] = '\0';
    strcpy(m_pszRegKey, pszAppRegKey);
    m_pszCourseTitle[0] = '\0';
}

CCourseLoader::~CCourseLoader()
{
    InvalidateDeviceObjects();
    DeleteDeviceObjects();
    m_pd3dDevice = NULL;
}

//loads a specified course file; should be called after InitDeviceObjects()
HRESULT CCourseLoader::LoadCourseFile(const char* pszPathname,
                                      const char* pszFilename,
                                      CItemManager* pItemManager,
                                      CAbstractGoal** ppGoal)
{
    _ASSERT(pszPathname && pszFilename && m_pd3dDevice);

    //free up any previous values
    InvalidateDeviceObjects();
    DeleteDeviceObjects();

    //store the file name
    strcpy(m_pszFilename, pszFilename);

    //read the high score for the new course
    ReadHighScore();

    //open the file
    char buf[512];
    strcpy(buf, pszPathname);
    strcat(buf, "\\");
    strcat(buf, pszFilename);
    ifstream infile(buf);
    if(!infile)
        return E_FAIL;

    //////////////////////////////////////////////////////
    //Read in the course title
    //////////////////////////////////////////////////////
    infile.getline(m_pszCourseTitle, 256);
    
    //////////////////////////////////////////////////////
    //Read in the terrain model
    //////////////////////////////////////////////////////
    //read in the terrain file name
    strcpy(buf, pszPathname);
    strcat(buf, "\\");
    infile >> (buf + strlen(buf));

    // Load the terrain file if there is one specified
    if(strcmp((buf + strlen(pszPathname) + 1), "0") != 0)
    {
        m_pTerrain = new CD3DMesh();
        if( FAILED( m_pTerrain->Create( m_pd3dDevice, _T(buf) ) ) )
        {
            infile.close();
            return E_FAIL;
        }
    }
    else
    {
        //use the course mesh as the terrain if terrain not specified
        m_pTerrain = NULL;
    }

    //////////////////////////////////////////////////////
    //read in the course mesh model
    //////////////////////////////////////////////////////
    //read in the course mesh file name
    strcpy(buf, pszPathname);
    strcat(buf, "\\");
    infile >> (buf + strlen(buf));

    // Load the course mesh if there is one (could be all landscape meshes)
    if(strcmp((buf + strlen(pszPathname) + 1), "0") != 0)
    {
        m_pCourseMesh = new CD3DMesh();
        if( FAILED( m_pCourseMesh->Create( m_pd3dDevice, _T(buf) ) ) )
        {
            infile.close();
            return E_FAIL;
        }
    }
    else
    {
        m_pCourseMesh = NULL;
    }

    //////////////////////////////////////////////////////
    //Read in the skybox model
    //////////////////////////////////////////////////////
    //read in the skybox file name
    infile >> (buf + strlen(pszPathname) + 1);

    //Load the skybox model
    m_pSkyBox = new CD3DMesh();
    if( FAILED( m_pSkyBox->Create( m_pd3dDevice, _T(buf) ) ) )
    {
        infile.close();
        return E_FAIL;
    }

    //////////////////////////////////////////////////////
    //Read in the snake position/direction information
    //////////////////////////////////////////////////////
    infile >> m_iSnakeStarts;
    m_pSnakeStarts   = m_iSnakeStarts ? new D3DXVECTOR3[m_iSnakeStarts] : NULL;
    m_pSnakeMoveDirs = m_iSnakeStarts ? new D3DXVECTOR3[m_iSnakeStarts] : NULL;
    FLOAT fDist;
    for(UINT iSnakeInfo = 0; iSnakeInfo < m_iSnakeStarts; ++iSnakeInfo)
    {
        infile >> m_pSnakeStarts[iSnakeInfo].x
               >> m_pSnakeStarts[iSnakeInfo].y
               >> m_pSnakeStarts[iSnakeInfo].z;

        //place all snakes 1 unit above the ground at the specified position
        m_pSnakeStarts[iSnakeInfo].z = 200.0f;
        if(GetTerrainIntersect(m_pSnakeStarts[iSnakeInfo], D3DXVECTOR3(0.0f, 0.0f, -1.0f), &fDist))
            m_pSnakeStarts[iSnakeInfo].z = 200.0f - fDist + 1;

        infile >> m_pSnakeMoveDirs[iSnakeInfo].x
               >> m_pSnakeMoveDirs[iSnakeInfo].y
               >> m_pSnakeMoveDirs[iSnakeInfo].z;
        m_pSnakeMoveDirs[iSnakeInfo] /= D3DXVec3Length(&m_pSnakeMoveDirs[iSnakeInfo]);
    }

    //////////////////////////////////////////////////////
    //Read in the water information
    //////////////////////////////////////////////////////
    UINT iWaterPlanes;
    infile >> iWaterPlanes;
    if(iWaterPlanes)
    {
        D3DXVECTOR3 vWaterPos;
        D3DXVECTOR2 vWaterSize;
        UINT        iHorzDiv, iVertDiv;
        infile >> vWaterPos.x >> vWaterPos.y >> vWaterPos.z;
        infile >> vWaterSize.x >> vWaterSize.y;
        infile >> iHorzDiv >> iVertDiv;
        m_pWaterPlane = new CWaterPlane(vWaterPos, vWaterSize, iHorzDiv, iVertDiv);
        m_pWaterPlane->InitDeviceObjects(m_pd3dDevice);
    }

    //////////////////////////////////////////////////////
    //Read in the wall information
    //////////////////////////////////////////////////////
    
    //read in the wall vertices information
    infile >> m_iWallsSize;
    m_pWalls = m_iWallsSize ? new CWallSection[m_iWallsSize] : NULL;
    for(UINT iWall = 0; iWall < m_iWallsSize; ++iWall)
    {
        m_pWalls[iWall].LoadFromFile(infile);
    }

    //////////////////////////////////////////////////////
    //Read in the landscape mesh information
    //////////////////////////////////////////////////////
    infile >> m_iMeshObjectSize;
    m_pMeshObject = m_iMeshObjectSize ? new CD3DMesh[m_iMeshObjectSize] : NULL;
    for(UINT iMesh = 0; iMesh < m_iMeshObjectSize; ++iMesh)
    {
        infile >> (buf + strlen(pszPathname) + 1);
        if( FAILED( m_pMeshObject[iMesh].Create( m_pd3dDevice, _T(buf) ) ) )
        {
            infile.close();
            return E_FAIL;
        }
        m_pMeshObject[iMesh].SetFVF( m_pd3dDevice, D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX1 );
    }

    UINT iTemp;
    infile >> m_iMeshPlaceSize;
    m_pMeshPlace = m_iMeshPlaceSize ? new MESHPLACEMENTINFO[m_iMeshPlaceSize] : NULL;
    for(UINT iMeshPlace = 0; iMeshPlace < m_iMeshPlaceSize; ++iMeshPlace)
    {
        infile >> iTemp;
        _ASSERT(iTemp < m_iMeshObjectSize);
        m_pMeshPlace[iMeshPlace].pMesh = &m_pMeshObject[iTemp];
        infile >> m_pMeshPlace[iMeshPlace].vPos.x
               >> m_pMeshPlace[iMeshPlace].vPos.y
               >> m_pMeshPlace[iMeshPlace].vPos.z;
    }

    //////////////////////////////////////////////////////
    //Read in the snake items
    //////////////////////////////////////////////////////
    if(pItemManager)
    {
        pItemManager->InitDeviceObjects(m_pd3dDevice);
        if(FAILED( pItemManager->ReadItems(pszPathname, infile) ))
        {
            infile.close();
            return E_FAIL;
        }
        pItemManager->RestoreDeviceObjects();
        pItemManager->Reset(this);      //place the snake items
    }

    //////////////////////////////////////////////////////
    //Read the goal
    //////////////////////////////////////////////////////
    if(ppGoal)
    {
        infile >> buf;
        CGoalFactory goalfact;
        *ppGoal = goalfact.CreateFromFile(buf, infile);
        _ASSERT(*ppGoal);
    }

    infile.close();

    //restore the device objects
    return S_OK;
}

//returns the filename of the currently loaded course
TCHAR* CCourseLoader::GetFileName()
{
    return m_pszFilename;
}

//initializes the snake start positions
void CCourseLoader::ResetSnakePositions(CSnakeManager* pSnakeManager, BOOL bMultiplayer)
{
    _ASSERT(pSnakeManager && m_pSnakeStarts && m_pSnakeMoveDirs);

    //multiplayer positions are everything after the first one
    //single player is the first one
    if(bMultiplayer)
    {
        if(m_iSnakeStarts > 0)
            pSnakeManager->SetSnakePositions(m_iSnakeStarts - 1, m_pSnakeStarts + 1, m_pSnakeMoveDirs + 1);
        else    //initialize the default starting positions
            pSnakeManager->SetSnakePositions(0, m_pSnakeStarts + 1, m_pSnakeMoveDirs + 1);
    }
    else
    {
        pSnakeManager->SetSnakePositions(m_iSnakeStarts, m_pSnakeStarts, m_pSnakeMoveDirs);
    }
}

//resets the course to the initial state
void CCourseLoader::Reset()
{
}

//advances to next frame of animation
void CCourseLoader::FrameMove()
{
    if(m_pWaterPlane)
    {
        static UINT iDelayCnt = 0;
        ++iDelayCnt;
        if(iDelayCnt > 5)
        {
            m_pWaterPlane->FrameMove();
            iDelayCnt = 0;
        }
    }
}

//checks for any interaction between course and snake and updates the course
//used for water splash
void CCourseLoader::UpdateCourse(CSnake* pSnake)
{
    if(m_pWaterPlane)
        m_pWaterPlane->UpdateSplashes(pSnake);
}

//render the sky of the scene
void CCourseLoader::RenderSky()
{
    _ASSERT(m_pd3dDevice && m_pSkyBox);

    m_pd3dDevice->LightEnable( 0, FALSE );
    m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE);

    // Center view matrix for skybox and disable zbuffer
    D3DXMATRIX matWorld;
    D3DXMatrixTranslation(&matWorld, 0.0f, 0.0f, 40.0f);
    m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

    D3DXMatrixScaling(&matWorld, 40.0f, 40.0f, 40.0f);
    m_pd3dDevice->MultiplyTransform( D3DTS_WORLD, &matWorld );

    D3DXMatrixRotationX(&matWorld, -3.1415f / 2.0f);
    m_pd3dDevice->MultiplyTransform( D3DTS_WORLD, &matWorld );
    m_pd3dDevice->SetRenderState( D3DRS_ZENABLE, FALSE );
    // Some cards do not disable writing to Z when 
    // D3DRS_ZENABLE is FALSE. So do it explicitly
    m_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );

    // Render the skybox
    m_pSkyBox->Render( m_pd3dDevice );

    // Restore the render states
    D3DXMatrixIdentity(&matWorld);
    m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

    m_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );
    m_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE, TRUE);

    m_pd3dDevice->LightEnable( 0, TRUE );
    m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, TRUE);
}


//draw the course on the given Direct3DDevice
void CCourseLoader::Render(CD3DCamera* pCamera, CSnake* pSnake)
{
    _ASSERT(m_pd3dDevice);

    ///////////////////////////////////////////////////////
    // Render the walls : for DEBUG purposes only
    ///////////////////////////////////////////////////////
    /*m_pd3dDevice->SetTexture( 0, NULL );
    m_pd3dDevice->SetVertexShader( D3DFVF_CUSTOMVERTEX );
    for(UINT iWall = 0; iWall < m_iWallsSize; ++iWall)
    {
        m_pWalls[iWall].Render(m_pd3dDevice, g_pWallSect, 4);
    }*/

    ///////////////////////////////////////////////////////
    // Render the ground
    ///////////////////////////////////////////////////////
    if(m_pCourseMesh)
        m_pCourseMesh->Render( m_pd3dDevice );

    ///////////////////////////////////////////////////////
    // Render the landscape meshes
    ///////////////////////////////////////////////////////

    //calculate the distance from the camera to the point of interest
    D3DXVECTOR3 vEye;
    FLOAT       fCamDist = 0;
    if(pCamera)
    {
        vEye = pCamera->GetEyePt();
        D3DXVECTOR3 vLookAt = pCamera->GetLookatPt();
        vLookAt -= vEye;
        fCamDist = D3DXVec3Length(&vLookAt);
    }

    D3DXMATRIX  matWorld;
    D3DXVECTOR3 vDir;
    BOOL        bHit = FALSE;
    FLOAT       fDist;

    m_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
    m_pd3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
    m_pd3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1 );
    m_pd3dDevice->SetRenderState(D3DRS_ALPHATESTENABLE, TRUE);
    m_pd3dDevice->SetRenderState(D3DRS_ALPHAFUNC, D3DCMP_GREATER);
    m_pd3dDevice->SetRenderState(D3DRS_ALPHAREF, 15);   //account for filtering

    for(UINT iMeshPlace = 0; iMeshPlace < m_iMeshPlaceSize; ++iMeshPlace)
    {
        if(pCamera)
        {
            //if the camera point of view intersects one of the landscape meshes
            //make the landscape mesh transparent
            vEye = pCamera->GetEyePt() - m_pMeshPlace[iMeshPlace].vPos;
            vDir = pCamera->GetViewDir();
            if(SUCCEEDED( D3DXIntersect(m_pMeshPlace[iMeshPlace].pMesh->GetSysMemMesh(),
                                &vEye, &vDir, &bHit,
                                NULL, NULL, NULL, &fDist, NULL, NULL) ) && bHit &&
                                fDist < fCamDist)
            {
                //if diving and distance is greater than 4 then make transparent
                if(~pSnake->GetCurrentState() & CSnake::ACTION_DIVE || (fCamDist - fDist) > 7.0f)
                {
                    m_pMeshPlace[iMeshPlace].bTransparent = TRUE;
                    continue;
                }
            }
        }
        m_pMeshPlace[iMeshPlace].bTransparent = FALSE;

        D3DXMatrixTranslation(&matWorld,
                              m_pMeshPlace[iMeshPlace].vPos.x,
                              m_pMeshPlace[iMeshPlace].vPos.y,
                              m_pMeshPlace[iMeshPlace].vPos.z);
        m_pd3dDevice->MultiplyTransform( D3DTS_WORLD, &matWorld );

        m_pMeshPlace[iMeshPlace].pMesh->Render(m_pd3dDevice, CD3DMesh::NO_USE_MATALPHA);

        D3DXMatrixTranslation(&matWorld,
                              -m_pMeshPlace[iMeshPlace].vPos.x,
                              -m_pMeshPlace[iMeshPlace].vPos.y,
                              -m_pMeshPlace[iMeshPlace].vPos.z);
        m_pd3dDevice->MultiplyTransform( D3DTS_WORLD, &matWorld );
    }

    //render the transparent landscape meshes that are obscuring the view
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE );
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_MODULATE );
    for(UINT iMesh = 0; iMesh < m_iMeshPlaceSize; ++iMesh)
    {
        D3DXMatrixTranslation(&matWorld,
                              m_pMeshPlace[iMesh].vPos.x,
                              m_pMeshPlace[iMesh].vPos.y,
                              m_pMeshPlace[iMesh].vPos.z);
        m_pd3dDevice->MultiplyTransform( D3DTS_WORLD, &matWorld );

        if(m_pMeshPlace[iMesh].bTransparent)
            m_pMeshPlace[iMesh].pMesh->Render( m_pd3dDevice );

        D3DXMatrixTranslation(&matWorld,
                              -m_pMeshPlace[iMesh].vPos.x,
                              -m_pMeshPlace[iMesh].vPos.y,
                              -m_pMeshPlace[iMesh].vPos.z);
        m_pd3dDevice->MultiplyTransform( D3DTS_WORLD, &matWorld );
    }

    m_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
    m_pd3dDevice->SetRenderState(D3DRS_ALPHATESTENABLE, FALSE);

    ///////////////////////////////////////////////////////
    // Render the water plane
    ///////////////////////////////////////////////////////
    if(m_pWaterPlane)
        m_pWaterPlane->Render();
}

//returns the terrain mesh for determining position values in the course
CD3DMesh* CCourseLoader::GetTerrain()
{
    if(!m_pTerrain)
        return m_pCourseMesh;
    return m_pTerrain;
}

//finds the intersection of the ray with the terrain
BOOL CCourseLoader::GetTerrainIntersect(const D3DXVECTOR3 &vRayPos,
                                        const D3DXVECTOR3 &vRay, FLOAT* pfDist)
{
    _ASSERT(m_pTerrain || m_pCourseMesh);

    BOOL bHit;
    if(m_pTerrain)
    {
        if(SUCCEEDED( D3DXIntersect(m_pTerrain->GetSysMemMesh(),
                                    &vRayPos, &vRay, &bHit,
                                    NULL, NULL, NULL, pfDist, NULL, NULL) ) && bHit)
        {
            return TRUE;
        }
    }
    else
    {
        if(SUCCEEDED( D3DXIntersect(m_pCourseMesh->GetSysMemMesh(),
                                    &vRayPos, &vRay, &bHit,
                                    NULL, NULL, NULL, pfDist, NULL, NULL) ) && bHit)
        {
            return TRUE;
        }
    }

    return FALSE;
}

//returns whether the line segment intersects a wall or not and the wall segment
//that it intersects with
BOOL CCourseLoader::GetWallIntersect(const D3DXVECTOR3 &vFrom, const D3DXVECTOR3 &vTo,
                                     D3DXVECTOR3* pvWallFrom, D3DXVECTOR3* pvWallTo)
{
    _ASSERT(m_pWalls && m_iWallsSize);

    FLOAT       fDistSq;
    D3DXVECTOR3 vWallFrom, vWallTo;
    D3DXVECTOR3 vIntersect;
    BOOL        bFoundIntersect = FALSE;
    for(UINT iWall = 0; iWall < m_iWallsSize; ++iWall)
    {
        if(m_pWalls[iWall].GetWallIntersect(vFrom, vTo, &vIntersect, &vWallFrom, &vWallTo))
        {
            if(!bFoundIntersect)
            {
                //calculate the distance to the intersection
                D3DXVECTOR3 vDiff = vIntersect - vFrom;
                fDistSq = D3DXVec3LengthSq(&vDiff);
                if(pvWallFrom)
                    *pvWallFrom = vWallFrom;
                if(pvWallTo)
                    *pvWallTo   = vWallTo;
                bFoundIntersect = TRUE;
            }
            else
            {
                D3DXVECTOR3 vDiff      = vIntersect - vFrom;
                FLOAT       fNewDistSq = D3DXVec3LengthSq(&vDiff);

                //compare the distance to the new intersection with the last found one
                //and replace it if it is closer
                if(fNewDistSq < fDistSq)
                {
                    if(pvWallFrom)
                        *pvWallFrom = vWallFrom;
                    if(pvWallTo)
                        *pvWallTo   = vWallTo;
                }
            }
        }
    }        

    return bFoundIntersect;
}

//returns the distance squared to the nearest wall to the point given
//returns -1.0f if there is no nearest wall
FLOAT CCourseLoader::GetNearestWall(const D3DXVECTOR3 &vPoint, D3DXVECTOR3* pvWallFrom, D3DXVECTOR3* pvWallTo)
{
    _ASSERT(m_pWalls && m_iWallsSize);

    FLOAT fNearDist = -1.0f;
    FLOAT fDist;
    D3DXVECTOR3 vWallFrom;
    D3DXVECTOR3 vWallTo;
    for(UINT iWall = 0; iWall < m_iWallsSize; ++iWall)
    {
        fDist = m_pWalls[iWall].GetNearestWall(vPoint, &vWallFrom, &vWallTo);
        if(fNearDist == -1.0f || (fDist != -1.0f && fDist < fNearDist))
        {
            fNearDist = fDist;
            if(pvWallFrom)
                *pvWallFrom = vWallFrom;
            if(pvWallTo)
                *pvWallTo = vWallTo;
        }
    }
    return fNearDist;
}

//returns a new windows regions defining the wall outline
//region must be deleted by caller using DeleteObject()
HRGN CCourseLoader::CreateWallRegion()
{
    _ASSERT(m_pWalls && m_iWallsSize);
    return m_pWalls[0].CreateWallRegion();
}

//returns the title of the course
TCHAR* CCourseLoader::GetCourseTitle()
{
    return m_pszCourseTitle;
}

//returns the high score of this course and the name of the high score
ULONG  CCourseLoader::GetHighScore(ULONG *piHighScore, const TCHAR** ppszName)
{
    if(piHighScore)
        *piHighScore = m_iHighScore;
    if(ppszName)
        *ppszName = m_pszHighName;
    return m_iHighScore;
}

//updates the high score for this course
void  CCourseLoader::UpdateHighScore(ULONG iHighScore, const TCHAR* pszName)
{
    _ASSERT(pszName);

    strcpy(m_pszHighName, pszName);
    m_iHighScore = iHighScore;
    WriteHighScore();
}

HRESULT CCourseLoader::InitDeviceObjects( LPDIRECT3DDEVICE8 pd3dDevice )
{
    // Keep a local copy of the device
    m_pd3dDevice = pd3dDevice;
    return S_OK;
}


HRESULT CCourseLoader::RestoreDeviceObjects()
{
    if(!m_pd3dDevice || !m_pSkyBox)
        return S_OK;

    if(m_pTerrain)
        m_pTerrain->RestoreDeviceObjects( m_pd3dDevice );
    if(m_pCourseMesh)
        m_pCourseMesh->RestoreDeviceObjects( m_pd3dDevice );
    m_pSkyBox->RestoreDeviceObjects( m_pd3dDevice );
    if(m_pWaterPlane)
        m_pWaterPlane->RestoreDeviceObjects();

    //restore all of the loaded meshes
    if(m_pMeshObject)
    {
        for(UINT iMesh = 0; iMesh < m_iMeshObjectSize; ++iMesh)
        {
            m_pMeshObject[iMesh].RestoreDeviceObjects( m_pd3dDevice );
        }
    }

    return S_OK;
}

HRESULT CCourseLoader::InvalidateDeviceObjects()
{
    if(m_pTerrain)
        m_pTerrain->InvalidateDeviceObjects();
    if(m_pCourseMesh)
        m_pCourseMesh->InvalidateDeviceObjects();
    if(m_pSkyBox)
        m_pSkyBox->InvalidateDeviceObjects();
    if(m_pWaterPlane)
        m_pWaterPlane->InvalidateDeviceObjects();

    //invalidate all of the loaded meshes
    if(m_pMeshObject)
    {
        for(UINT iMesh = 0; iMesh < m_iMeshObjectSize; ++iMesh)
        {
            m_pMeshObject[iMesh].InvalidateDeviceObjects();
        }
    }

    return S_OK;
}

HRESULT CCourseLoader::DeleteDeviceObjects()
{
    if(m_pTerrain)
        m_pTerrain->Destroy();
    if(m_pCourseMesh)
        m_pCourseMesh->Destroy();
    if(m_pSkyBox)
        m_pSkyBox->Destroy();
    if(m_pWaterPlane)
        m_pWaterPlane->DeleteDeviceObjects();

    //delete all of the loaded meshes
    if(m_pMeshObject)
    {
        for(UINT iMesh = 0; iMesh < m_iMeshObjectSize; ++iMesh)
        {
            m_pMeshObject[iMesh].Destroy();
        }
        SAFE_DELETE_ARRAY( m_pMeshObject );
        m_iMeshObjectSize = 0;
    }

    SAFE_DELETE( m_pTerrain );
    SAFE_DELETE( m_pCourseMesh );
    SAFE_DELETE( m_pSkyBox );
    SAFE_DELETE( m_pWaterPlane );
    SAFE_DELETE_ARRAY( m_pWalls );
    m_iWallsSize = 0;

    SAFE_DELETE_ARRAY( m_pSnakeStarts );
    SAFE_DELETE_ARRAY( m_pSnakeMoveDirs );
    m_iSnakeStarts = 0;

    SAFE_DELETE_ARRAY( m_pMeshPlace );
    m_iMeshPlaceSize = 0;
    m_pszFilename[0] = '\0';
    return S_OK;
}

//reads the high score from the registry
void CCourseLoader::ReadHighScore()
{
    //if a course hasn't been loaded then don't read anything
    if(m_pszFilename[0] == '\0')
        return;

    char buf[512];
    strcpy(buf, m_pszRegKey);
    strcat(buf, "\\");
    strcat(buf, m_pszFilename);
    HKEY hkey;
    if( ERROR_SUCCESS == RegCreateKeyEx( HKEY_CURRENT_USER, buf, 
        0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hkey, NULL ) )
    {
        DXUtil_ReadStringRegKey( hkey, "HighName",  m_pszHighName, 256, "NONE");
        DXUtil_ReadIntRegKey( hkey, "HighScore", &m_iHighScore, 0L );
        RegCloseKey( hkey );
    }
}

//writes the high score to the registry
void CCourseLoader::WriteHighScore()
{
    //if a course hasn't been loaded then don't write anything
    if(m_pszFilename[0] == '\0')
        return;

    char buf[512];
    strcpy(buf, m_pszRegKey);
    strcat(buf, "\\");
    strcat(buf, m_pszFilename);
    HKEY hkey;
    if( ERROR_SUCCESS == RegCreateKeyEx( HKEY_CURRENT_USER, buf, 
        0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hkey, NULL ) )
    {
        DXUtil_WriteStringRegKey( hkey, "HighName", m_pszHighName);
        DXUtil_WriteIntRegKey( hkey, "HighScore", m_iHighScore);
        RegCloseKey( hkey );
    }
}