// CourseLoader.h: interface for the CCourseLoader class.
//
//////////////////////////////////////////////////////////////////////
#pragma once

#include "d3dfile.h"
#include "d3dutil.h"
#include "dxutil.h"
#include "Lofter.h"
#include "ItemManager.h"
#include "SnakeManager.h"
#include "WallSection.h"
#include "AbstractGoal.h"
#include "WaterPlane.h"
#include <D3dx8math.h>

class CCourseLoader
{
public:
    CCourseLoader(const char* pszAppRegKey);
    ~CCourseLoader();

    //loads a specified course file
    HRESULT LoadCourseFile(const char* pszPathname, const char* pszFilename,
                           CItemManager* pItemManager = NULL,
                           CAbstractGoal** ppGoal = NULL);

    //initializes the snake start positions
    void ResetSnakePositions(CSnakeManager* pSnakeManager, BOOL bMultiplayer = FALSE);

    //returns the filename of the currently loaded course
    TCHAR* GetFileName();

    //returns the terrain mesh
    CD3DMesh* GetTerrain();

    //finds the intersection of the ray with the terrain
    BOOL GetTerrainIntersect(const D3DXVECTOR3 &vRayPos, const D3DXVECTOR3 &vRay,
                             FLOAT* pfDist = NULL);

    //returns whether the line segment intersects a wall or not
    BOOL GetWallIntersect(const D3DXVECTOR3 &vFrom, const D3DXVECTOR3 &vTo,
                          D3DXVECTOR3* pvWallFrom = NULL, D3DXVECTOR3* pvWallTo = NULL);

    //returns the distance squared to the nearest wall to the point given
    FLOAT GetNearestWall(const D3DXVECTOR3 &vPoint, D3DXVECTOR3* pvWallFrom = NULL, D3DXVECTOR3* pvWallTo = NULL);

    //returns a new windows regions defining the wall outline
    HRGN CreateWallRegion();

    //returns the title of the course
    TCHAR* GetCourseTitle();

    //returns the high score of this course
    ULONG  GetHighScore(ULONG *piHighScore = NULL, const TCHAR** ppszName = NULL);

    //updates the high score for this course
    void   UpdateHighScore(ULONG iHighScore, const TCHAR* pszName);

    //resets the course to the initial state
    void Reset();

    //advances to next frame of animation
    void FrameMove();

    //checks for any interaction between course and snake and updates the course
    //used for water splash
    void UpdateCourse(CSnake* pSnake);

    //render the skybox of the course
    void RenderSky();

    //main render function
    void Render(CD3DCamera* pCamera = NULL, CSnake* pSnake = NULL);

    // Initializing and destroying device-dependent objects
    HRESULT InitDeviceObjects( LPDIRECT3DDEVICE8 pd3dDevice );
    HRESULT RestoreDeviceObjects();
    HRESULT InvalidateDeviceObjects();
    HRESULT DeleteDeviceObjects();
protected:

    //reads the high score from the registry
    void    ReadHighScore();

    //writes the high score to the registry
    void    WriteHighScore();

private:
    LPDIRECT3DDEVICE8           m_pd3dDevice;           // A D3DDevice used for rendering
    TCHAR                       m_pszRegKey[128];
    TCHAR                       m_pszFilename[128];

    TCHAR                       m_pszCourseTitle[256];  // title is a maximum of 256 chars
    CD3DMesh*                   m_pTerrain;             // mesh for the terrain for intersections
    CD3DMesh*                   m_pCourseMesh;          // mesh that is actually drawn
    CD3DMesh*                   m_pSkyBox;              // mesh for the sky
    CWaterPlane*                m_pWaterPlane;          // the water plane

    CWallSection*               m_pWalls;
    UINT                        m_iWallsSize;
    
    CD3DMesh*                   m_pMeshObject;
    UINT                        m_iMeshObjectSize;

    D3DXVECTOR3*                m_pSnakeStarts;         //contains a list of snake start positions (single & multiplayer)
    D3DXVECTOR3*                m_pSnakeMoveDirs;       //contains a list of snake move directions (single & multiplayer)  
    UINT                        m_iSnakeStarts;

    struct MESHPLACEMENTINFO
    {
        CD3DMesh*       pMesh;
        D3DXVECTOR3     vPos;
        BOOL            bTransparent;
    };

    MESHPLACEMENTINFO*          m_pMeshPlace;
    UINT                        m_iMeshPlaceSize;

    TCHAR                       m_pszHighName[256];
    ULONG                       m_iHighScore;
};