// GameMath.cpp: implementation of the CGameMath class.
//
//////////////////////////////////////////////////////////////////////

#include "GameMath.h"
#include <crtdbg.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CGameMath::CGameMath()
{

}

CGameMath::~CGameMath()
{

}

//returns whether there is an intersects in the XY plane and the
//returns the distance in the Z direction at the intersection
BOOL CGameMath::GetLineIntersect(const D3DXVECTOR3& vFrom1, const D3DXVECTOR3& vTo1,
                                 const D3DXVECTOR3& vFrom2, const D3DXVECTOR3& vTo2,
                                 D3DXVECTOR3* pIntersect, FLOAT* fZdelta)
{
    //find the slope and intercept for the first line
    FLOAT fXDeltaXY1 = vTo1.x - vFrom1.x;
    FLOAT fSlopeXY1 = (vTo1.y - vFrom1.y) / fXDeltaXY1;
    FLOAT fInterceptXY1 = vFrom1.y - fSlopeXY1 * vFrom1.x;

    //find the slope and intercept for the second line
    FLOAT fXDeltaXY2 = vTo2.x - vFrom2.x;
    FLOAT fSlopeXY2 = (vTo2.y - vFrom2.y) / fXDeltaXY2;
    FLOAT fInterceptXY2 = vFrom2.y - fSlopeXY2 * vFrom2.x;

    //same slope so no intersection
    if(fSlopeXY1 == fSlopeXY2)
        return FALSE;

    FLOAT fIntersectX;
    FLOAT fIntersectY;
    if(fXDeltaXY1 && fXDeltaXY2)    //both lines have a slope
    {
        fIntersectX = (fInterceptXY2 - fInterceptXY1) / (fSlopeXY1 - fSlopeXY2);
        fIntersectY = fSlopeXY1 * fIntersectX + fInterceptXY1;
    }
    else if(!fXDeltaXY1)    //first line has no slope
    {
        fIntersectX = vTo1.x;
        fIntersectY = fSlopeXY2 * fIntersectX + fInterceptXY2;
    }
    else if(!fXDeltaXY2)    //second line has no slope
    {
        fIntersectX = vTo2.x;
        fIntersectY = fSlopeXY1 * fIntersectX + fInterceptXY1;
    }
    else    //both cannot have no slope or else there is no intersection
    {
        _ASSERT(FALSE);
    }

    //account for roundoff errors
    if(((fIntersectX >= vTo2.x - 1e-3 && fIntersectX <= vFrom2.x + 1e-3) ||
        (fIntersectX <= vTo2.x + 1e-3 && fIntersectX >= vFrom2.x - 1e-3)) &&
       ((fIntersectY >= vTo2.y - 1e-1 && fIntersectY <= vFrom2.y + 1e-1) ||
        (fIntersectY <= vTo2.y + 1e-1 && fIntersectY >= vFrom2.y - 1e-1)) &&
       ((fIntersectX >= vFrom1.x - 1e-3 && fIntersectX <= vTo1.x + 1e-3) ||
        (fIntersectX <= vFrom1.x + 1e-3 && fIntersectX >= vTo1.x - 1e-3)) &&
       ((fIntersectY >= vFrom1.y - 1e-1 && fIntersectY <= vTo1.y + 1e-1) ||
        (fIntersectY <= vFrom1.y + 1e-1 && fIntersectY >= vTo1.y - 1e-1)))
    {
        if(pIntersect)
            *pIntersect = D3DXVECTOR3(fIntersectX, fIntersectY, 0.0f);

        if(fZdelta)
        {
            D3DXVECTOR3 vDiff1 = vTo1 - vFrom1;
            D3DXVECTOR3 vDiff2 = vTo2 - vFrom2;

            vDiff1 /= D3DXVec3Length(&vDiff1);
            vDiff2 /= D3DXVec3Length(&vDiff2);

            if(fIntersectX - vFrom1.x)
                vDiff1 *= (fIntersectX - vFrom1.x) / vDiff1.x;
            else
                vDiff1 *= (fIntersectY - vFrom1.y) / vDiff1.y;

            if(fIntersectX - vFrom2.x)
                vDiff2 *= (fIntersectX - vFrom2.x) / vDiff2.x;
            else
                vDiff2 *= (fIntersectY - vFrom2.y) / vDiff2.y;

            vDiff1 += vFrom1;
            vDiff2 += vFrom2;

            *fZdelta = vDiff1.z - vDiff2.z;
        }
        return TRUE;
    }

    return FALSE;
}

//returns whether there is a perpendicular to the line segment that goes
//through the given point, and returns the point on the line segment,
//and the percent between the from and the two points of the intersection
BOOL CGameMath::GetPointPerp(const D3DXVECTOR3& vFrom, const D3DXVECTOR3& vTo,
                             const D3DXVECTOR3& vPoint, D3DXVECTOR3* pIntersectPt,
                             FLOAT* pfPercent)
{
    D3DXVECTOR3 vDiff1 = vPoint - vFrom;
    D3DXVECTOR3 vDiff2 = vTo - vFrom;

    //calculate the projection of vDiff1 onto vDiff2
    FLOAT fOffset = vDiff1.x * vDiff2.x + vDiff1.y * vDiff2.y;
    fOffset /= vDiff2.x * vDiff2.x + vDiff2.y * vDiff2.y;

    //check whether it is within the line segment
    if(fOffset < 0.0f || fOffset > 1.0f)
        return FALSE;
    
    if(pfPercent)
        *pfPercent = fOffset;

    //find the intersection
    if(pIntersectPt)
        *pIntersectPt = vFrom + fOffset * vDiff2;
    
    return TRUE;
}

//calculates the camera view angle based on the direction vector
FLOAT CGameMath::CalculateCameraAngle(const D3DXVECTOR3& vDir)
{
    //find the current angle of the snake
    FLOAT curangle;
    if(vDir.x)
    {
        curangle = (FLOAT)atan2(vDir.y, vDir.x);
        if(curangle < 0)
            curangle += 2.0f * 3.1415f;
    }
    else
    {
        curangle = ((vDir.y > 0) ? 3.1415f / 2.0f : 3.0f * 3.1415f / 2.0f);
    }

    return curangle;
}
