// AbstractItem.cpp: implementation of the CItem class.
//
//////////////////////////////////////////////////////////////////////

#include "ItemBehaviorFactory.h"
#include "AbstractItemBehavior.h"
#include "CourseLoader.h"
#include "Item.h"
#include <crtdbg.h>     //for _ASSERT

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CItem::CItem() : m_pMesh(NULL),
                 m_fElevation(1.0f),
                 m_vPos(0.0f, 0.0f, 0.0f),
                 m_vRotation(0.0f, 0.0f, 0.0f),
                 m_bValid(FALSE),
                 m_bVisible(FALSE),
                 m_ppBehaviors(NULL),
                 m_iNumBehaviors(0)
{
}

CItem::~CItem()
{
    for(UINT i = 0; i < m_iNumBehaviors; ++i)
        SAFE_DELETE( m_ppBehaviors[i] );
    SAFE_DELETE_ARRAY( m_ppBehaviors );
}

//reads the behaviors for the item from the specified file
void CItem::ReadItemBehaviors(UINT iNumBehaviors, ifstream& infile)
{
    SAFE_DELETE_ARRAY( m_ppBehaviors );

    CItemBehaviorFactory fact;
    m_ppBehaviors   = new CAbstractItemBehavior*[iNumBehaviors];
    m_iNumBehaviors = iNumBehaviors;
    char buf[512];
    for(UINT i = 0; i < m_iNumBehaviors; ++i)
    {
        infile >> buf;
        m_ppBehaviors[i] = fact.CreateFromFile(buf, infile);
        _ASSERT(m_ppBehaviors[i]);
    }
}

//render the snake item
void CItem::Render(LPDIRECT3DDEVICE8 pd3dDevice)
{
    _ASSERT(pd3dDevice);

    if(!m_pMesh)
        return;

    D3DXMATRIX matWorld1, matWorld2, matWorld3;

    //translate to the correct position
    D3DXMatrixTranslation(&matWorld1, m_vPos.x, m_vPos.y, m_vPos.z);
    
    //perform all of the rotations
    D3DXMatrixRotationX(&matWorld2, m_vRotation.x);
    D3DXMatrixMultiply(&matWorld3, &matWorld2, &matWorld1);
    D3DXMatrixRotationY(&matWorld2, m_vRotation.y);
    D3DXMatrixMultiply(&matWorld1, &matWorld2, &matWorld3);
    D3DXMatrixRotationZ(&matWorld2, m_vRotation.z);
    D3DXMatrixMultiply(&matWorld3, &matWorld2, &matWorld1);
    pd3dDevice->MultiplyTransform( D3DTS_WORLD, &matWorld3 );

    //render the mesh
    m_pMesh->Render( pd3dDevice );

    //revert to previous world matrix
    D3DXMatrixInverse(&matWorld1, NULL, &matWorld3);
    pd3dDevice->MultiplyTransform( D3DTS_WORLD, &matWorld1 );
}

//set the position of this item
//returns whether there was a terrain intersect or not
BOOL CItem::SetPos(const D3DXVECTOR3& vPos, CCourseLoader* pCourse)
{
    m_vPos = vPos;

    //check for terrain intersect down from the item
    FLOAT fDist;
    if(pCourse->GetTerrainIntersect(m_vPos, D3DXVECTOR3(0.0f, 0.0f, -1.0f), &fDist))
    {
        m_vPos.z += m_fElevation - fDist;
    }
    else if(pCourse->GetTerrainIntersect(m_vPos, D3DXVECTOR3(0.0f, 0.0f, 1.0f), &fDist))
    {
        m_vPos.z += m_fElevation + fDist;
    }
    else
    {
        //Could not find terrain intersection
        m_vPos.z = 0.0f;
        return FALSE;
    }
    return TRUE;
}

//invalidate this device to turn off collision detection
//and remove from rendering
void CItem::Invalidate()
{
    m_bVisible = FALSE;
    m_bValid   = FALSE;
}

//set the mesh to render with
void CItem::SetMesh(CD3DMesh* pMesh)
{
    _ASSERT(pMesh);
    m_pMesh    = pMesh;
    Reset();
}

//returns whether this item is visible
BOOL CItem::IsVisible()
{
    return m_bVisible;
}

//returns whether this item should be collision detected
BOOL CItem::IsValid()
{
    return m_bValid;
}

//////////////////////////////////////////////////////////////////////
// Item Behavior Functions
//////////////////////////////////////////////////////////////////////

//reset this item to the original state: visible, valid
void CItem::Reset(CCourseLoader* pCourseLoader)
{
    m_bVisible = TRUE;
    m_bValid   = TRUE;

    for(UINT i = 0; i < m_iNumBehaviors; ++i)
    {
        m_ppBehaviors[i]->Reset(*this, pCourseLoader);
    }
}

//move to the next frame of the animation
//ppSnakeVect is a NULL terminates array of CSnake*
void CItem::FrameMove(CCourseLoader* pCourse,
                      CItem* pItem, UINT iItemSize, CSnakeManager* pSnakeManager)
{
    for(UINT i = 0; i < m_iNumBehaviors; ++i)
    {
        m_ppBehaviors[i]->FrameMove(*this, pCourse, pItem, iItemSize, pSnakeManager);
    }
}

//checks whether the snake is in the bounds of this item
BOOL CItem::CheckInBounds(CSnake* pSnake)
{
    _ASSERT(pSnake);
    for(UINT i = 0; i < m_iNumBehaviors; ++i)
    {
        if(m_ppBehaviors[i]->CheckInBounds(pSnake, this))
            return TRUE;
    }
    return FALSE;
}

//performs an behavior associated with the item on the given snake
void CItem::ModifySnake(CSnake* pSnake)
{
    for(UINT i = 0; i < m_iNumBehaviors; ++i)
    {
        if(m_ppBehaviors[i]->ModifySnake(pSnake))
            Invalidate();
    }
}
