// ItemManager.cpp: implementation of the CItemManager class.
//
//////////////////////////////////////////////////////////////////////

#include "ItemManager.h"
#include "CourseLoader.h"
#include <crtdbg.h>     //for _ASSERT

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CItemManager::CItemManager() : m_pd3dDevice(NULL),
                               m_iMeshObjectSize(0),
                               m_pMeshObject(NULL),
                               m_iItemSize(0),
                               m_pItem(NULL)
{
}

CItemManager::~CItemManager()
{
    InvalidateDeviceObjects();
    DeleteDeviceObjects();
    m_pd3dDevice = NULL;
}


//reads the snake items from the given file
HRESULT CItemManager::ReadItems(const char* pszPathname, ifstream& infile)
{
    _ASSERT(pszPathname && m_pd3dDevice);

    //free up any previous values
    InvalidateDeviceObjects();
    DeleteDeviceObjects();

    char buf[512];
    strcpy(buf, pszPathname);
    strcat(buf, "\\");

    //////////////////////////////////////////////////////
    //Read in the mesh object information
    //////////////////////////////////////////////////////
    infile >> m_iMeshObjectSize;
    m_pMeshObject = m_iMeshObjectSize ? new CD3DMesh[m_iMeshObjectSize] : NULL;
    for(UINT iMesh = 0; iMesh < m_iMeshObjectSize; ++iMesh)
    {
        infile >> (buf + strlen(pszPathname) + 1);
        if( FAILED( m_pMeshObject[iMesh].Create( m_pd3dDevice, _T(buf) ) ) )
            return E_FAIL;
        m_pMeshObject[iMesh].SetFVF( m_pd3dDevice, D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX1 );
    }

    //////////////////////////////////////////////////////
    //Read in the snake items
    //////////////////////////////////////////////////////
    int               iMeshIndex;
    UINT              iNumBehaviors;
    infile >> m_iItemSize;
    m_pItem = m_iItemSize ? new CItem[m_iItemSize] : NULL;
    for(UINT iItem = 0; iItem < m_iItemSize; ++iItem)
    {
        infile >> iMeshIndex >> iNumBehaviors;
        m_pItem[iItem].ReadItemBehaviors(iNumBehaviors, infile);

        if(iMeshIndex < 0)
            continue;

        _ASSERT((UINT)iMeshIndex < m_iMeshObjectSize);
        m_pItem[iItem].SetMesh(&m_pMeshObject[iMeshIndex]);
        
    }

    return S_OK;
}

//positions the items contained in the manager into the course
void CItemManager::PositionItems(CCourseLoader* pCourse)
{
    _ASSERT(pCourse);

    HRGN hRegion = pCourse->CreateWallRegion();
    RECT rgnRect;
    GetRgnBox(hRegion, &rgnRect);

    FLOAT fNearDist;
    //read in the snake item information and initialize positions
    for(UINT iItem = 0; iItem < m_iItemSize; ++iItem)
    {
        //find a random position in the region
        D3DXVECTOR3 vPos;
        vPos.z = 1000.0f;
        do
        {
            vPos.x = (float)((rgnRect.right - rgnRect.left) * rand() / RAND_MAX + rgnRect.left);
            vPos.y = (float)((rgnRect.bottom - rgnRect.top) * rand() / RAND_MAX + rgnRect.top);
            fNearDist = pCourse->GetNearestWall(vPos);
            
            //set the item in the course at the item specified height
        }
        while(!PtInRegion(hRegion, (int)vPos.x, (int)vPos.y) ||
                fNearDist < 9.0f || fNearDist == -1.0 ||
                !m_pItem[iItem].SetPos(vPos, pCourse));
    }

    DeleteObject(hRegion);
}

//resets all of the snake items to their initial states
void CItemManager::Reset(CCourseLoader* pCourse)
{
    if(pCourse)
        PositionItems(pCourse);

    for(UINT iItem = 0; iItem < m_iItemSize; ++iItem)
    {
        m_pItem[iItem].Reset(pCourse);
    }
}

//advance to the next frame of animation for the snake items
void CItemManager::FrameMove(CCourseLoader* pCourse, CSnakeManager* pSnakeManager)
{
    for(UINT iItem = 0; iItem < m_iItemSize; ++iItem)
    {
        if(m_pItem[iItem].IsVisible())
            m_pItem[iItem].FrameMove(pCourse, m_pItem, m_iItemSize, pSnakeManager);
    }
}


//returns the item specified by the ID
CItem* CItemManager::GetItem(UINT iItemID)
{
    if(iItemID < m_iItemSize)
        return &m_pItem[iItemID];
    return NULL;
}

//returns whether the given snake intersects with an item and returns
//the given item which the caller can invalidates, uses the lastitem
//returned by GetItemIntersect to start search
CItem* CItemManager::GetItemIntersect(CSnake* pSnake, CItem* pStartItem)
{
    if(!m_pItem || !m_iItemSize)
        return NULL;

    _ASSERT(pSnake);

    if(!pStartItem)
        pStartItem = m_pItem;
    else
        ++pStartItem;

    while(pStartItem < m_pItem + m_iItemSize)
    {
        //skip invalid items
        if(!pStartItem->IsValid() || !pStartItem->IsVisible())
        {
            ++pStartItem;
            continue;
        }

        if(pStartItem->CheckInBounds(pSnake))
            return pStartItem;

        ++pStartItem;
    }
    return NULL;
}

void CItemManager::Render()
{
    for(UINT iItem = 0; iItem < m_iItemSize; ++iItem)
    {
        if(m_pItem[iItem].IsVisible())
            m_pItem[iItem].Render(m_pd3dDevice);
    }
}

HRESULT CItemManager::InitDeviceObjects( LPDIRECT3DDEVICE8 pd3dDevice )
{
    // Keep a local copy of the device
    m_pd3dDevice = pd3dDevice;

    return S_OK;
}

HRESULT CItemManager::RestoreDeviceObjects()
{
    if(!m_pMeshObject)
        return S_OK;

    //restore all of the loaded meshes
    for(UINT iMesh = 0; iMesh < m_iMeshObjectSize; ++iMesh)
    {
        m_pMeshObject[iMesh].RestoreDeviceObjects( m_pd3dDevice );
    }

    return S_OK;
}

HRESULT CItemManager::InvalidateDeviceObjects()
{
    //invalidate all of the loaded meshes
    if(!m_pMeshObject)
        return S_OK;
        
    for(UINT iMesh = 0; iMesh < m_iMeshObjectSize; ++iMesh)
    {
        m_pMeshObject[iMesh].InvalidateDeviceObjects();
    }

    return S_OK;
}

HRESULT CItemManager::DeleteDeviceObjects()
{
    //restore all of the loaded meshes
    if(m_pMeshObject)
    {
        for(UINT iMesh = 0; iMesh < m_iMeshObjectSize; ++iMesh)
        {
            m_pMeshObject[iMesh].Destroy();
        }
        SAFE_DELETE_ARRAY( m_pMeshObject );
        m_iMeshObjectSize = 0;
    }

    if(m_pItem)
    {
        SAFE_DELETE_ARRAY( m_pItem );
        m_iItemSize = 0;
    }

    return S_OK;
}
