// Lofter.cpp: implementation of the CLofter class.
//
//////////////////////////////////////////////////////////////////////

#include "Lofter.h"
#include <crtdbg.h>     //for _ASSERT

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CLofter::CLofter()
{

}

CLofter::~CLofter()
{

}

//draw a lofted section that is beveled based on a path and a section outline
//vertarray is an array that holds:
//      [vertfrom] [vertto] (bevelfrom) (bevelto) (scalefrom) (scaleto)
//depending on the flags passed in
//pSect is the cross section to loft along the path specified in clockwise winding
CUSTOMVERTEX* CLofter::LoftBeveledSection(const D3DXVECTOR3* pVertArray,
                                          const VECTOR2TEX*  pSect,
                                          int                iSectSize,
                                          CUSTOMVERTEX*      pVertices,
                                          D3DCOLOR           color,
                                          DWORD              dwFlags)
{
    _ASSERT(pVertArray && pSect && iSectSize && pVertices);

    D3DXVECTOR3 vertFrom = pVertArray[0];
    D3DXVECTOR3 vertTo   = pVertArray[1];
    pVertArray += 2;
    
    //calculate the loft vector
    D3DXVECTOR3 delta = vertTo - vertFrom;
    if(fabs(delta.x) < 1e-5f && fabs(delta.y) < 1e-5f)
        return pVertices;

    //normalize the delta vector in the xy plane
    delta /= D3DXVec3Length(&delta); //(FLOAT)sqrt(delta.x * delta.x + delta.y * delta.y);

    //find the right vector in the xy plane
    D3DXVECTOR3 right = D3DXVECTOR3(delta.y, -delta.x, 0.0f);
    right /= D3DXVec3Length(&right);

    //find the up vector for this vector
    D3DXVECTOR3 up;
    if(delta.z < 0)  //sloping down
        up = D3DXVECTOR3(delta.x, delta.y, (FLOAT)-sqrt(delta.x*delta.x + delta.y*delta.y) / delta.z);
    else if(delta.z > 0)    //sloping up
        up = D3DXVECTOR3(-delta.x, -delta.y, (FLOAT)sqrt(delta.x*delta.x + delta.y*delta.y) / delta.z);
    else
        up = D3DXVECTOR3(0.0f, 0.0f, 1.0f);
    up /= D3DXVec3Length(&up);

    //////////////////////////////////////////////////////////////////////
    // Bevel Calculations
    //////////////////////////////////////////////////////////////////////

    //deal with the from bevel: find the right vector for the bevel
    D3DXVECTOR3 bevelFromUp;
    D3DXVECTOR3 bevelFromRight;
    if(dwFlags & LOFT_NOBEVELFROM)
    {
        bevelFromRight = right;
        bevelFromUp    = up;
    }
    else
    {
        //calculate the bevel vector
        D3DXVECTOR3 deltaBevel;
        deltaBevel = *pVertArray - vertFrom;

        //normalize the bevel vector
        FLOAT fBevLen = D3DXVec3Length(&deltaBevel);

        //calculate the bevel's right vector
        if(!fBevLen)
        {
            bevelFromRight = right;
            bevelFromUp    = up;
        }
        else
        {
            CalcBevelUpRight(&bevelFromUp, &bevelFromRight,
                             delta, deltaBevel / fBevLen, right, up);
        }

        ++pVertArray;
    }

    //deal with the to bevel: find the right vector for the bevel
    D3DXVECTOR3 bevelToUp;
    D3DXVECTOR3 bevelToRight;
    if(dwFlags & LOFT_NOBEVELTO)
    {
        bevelToRight = right;
        bevelToUp    = up;
    }
    else
    {
        D3DXVECTOR3 deltaBevel;

        //calculate the bevel vector
        deltaBevel = *pVertArray - vertTo;

        //normalize the bevel vector
        FLOAT fBevLen = D3DXVec3Length(&deltaBevel); //(FLOAT) sqrt(deltaBevel.x * deltaBevel.x + deltaBevel.y * deltaBevel.y);;

        //calculate the bevel's right vector
        if(!fBevLen)
        {
            bevelToRight = right;
            bevelToUp    = up;
        }
        else
        {
            //use the up vector from the next section in order to prevent
            //holes when the cross product is zero, but there is a bevel present
            if(deltaBevel.z < 0)  //sloping down
                up = D3DXVECTOR3(deltaBevel.x, deltaBevel.y, (FLOAT)-sqrt(deltaBevel.x*deltaBevel.x + deltaBevel.y*deltaBevel.y) / deltaBevel.z);
            else if(deltaBevel.z > 0)    //sloping up
                up = D3DXVECTOR3(-deltaBevel.x, -deltaBevel.y, (FLOAT)sqrt(deltaBevel.x*deltaBevel.x + deltaBevel.y*deltaBevel.y) / deltaBevel.z);
            else
                up = D3DXVECTOR3(0.0f, 0.0f, 1.0f);
            up /= D3DXVec3Length(&up);

            CalcBevelUpRight(&bevelToUp, &bevelToRight,
                             -delta, deltaBevel / fBevLen, right, up);
        }

        ++pVertArray;
    }

    //////////////////////////////////////////////////////////////////////
    // Scaling Retrieval
    //////////////////////////////////////////////////////////////////////
    D3DXVECTOR3 scaleFrom;
    if(dwFlags & LOFT_SCALEFROM)
    {
        scaleFrom = pVertArray[0];
        ++pVertArray;
    }
    else
    {
        scaleFrom = D3DXVECTOR3(1.0f, 1.0f, 1.0f);
    }

    D3DXVECTOR3 scaleTo;
    if(dwFlags & LOFT_SCALETO)
    {
        scaleTo = pVertArray[0];
        ++pVertArray;
    }
    else
    {
        scaleTo = D3DXVECTOR3(1.0f, 1.0f, 1.0f);
    }

    //////////////////////////////////////////////////////////////////////
    // Normal Calculations
    //////////////////////////////////////////////////////////////////////

    //calculate the relative normals for each vertex being drawn
    D3DXVECTOR2* pNormals = new D3DXVECTOR2[iSectSize];
    for(int index = 0; index < iSectSize; ++index)
    {
        FLOAT dxsect = pSect[index].x - pSect[(index + 1) % iSectSize].x;
        FLOAT dysect = pSect[index].y - pSect[(index + 1) % iSectSize].y;
        FLOAT ilength = (FLOAT) sqrt(dxsect*dxsect + dysect*dysect);
        dxsect /= ilength;
        dysect /= ilength;

        FLOAT dxsect2;
        FLOAT dysect2;

        //if it is the index is zero, then the previous index is the next to last index in pSect
        dxsect2 = pSect[index].x - pSect[(index + iSectSize - (index ? 1 : 2)) % iSectSize].x;
        dysect2 = pSect[index].y - pSect[(index + iSectSize - (index ? 1 : 2)) % iSectSize].y;
        ilength = (FLOAT) sqrt(dxsect2*dxsect2 + dysect2*dysect2);
        dxsect2 /= ilength;
        dysect2 /= ilength;

        //add the two vectors going into the point to obtain the relative normal
        dxsect += dxsect2;
        dysect += dysect2;
        ilength = (FLOAT) sqrt(dxsect*dxsect + dysect*dysect);
        dxsect /= ilength;
        dysect /= ilength;

        pNormals[index].x = dxsect;
        pNormals[index].y = dysect;
    }

    //////////////////////////////////////////////////////////////////////
    // Vertex Generation
    //////////////////////////////////////////////////////////////////////

    if(dwFlags & LOFT_INDEXED)
    {
        if(dwFlags & LOFT_ENDFROM)
        {
            //set the FROM vertices
            for(int i = 0; i < iSectSize; ++i)
            {
                //calculate the first vertex
                //set the texture mapping information
                pVertices->color = color;
                pVertices->tu = pSect[i].tu;
                pVertices->tv = 0.0f;
                CalcSectionVertex(pVertices, vertFrom, pNormals[i], pSect[i],
                                  bevelFromUp, bevelFromRight, scaleFrom);
                ++pVertices;
            }
        }

        //set the TO vertices
        for(int i = 0; i < iSectSize; ++i)
        {
            //calculate the second vertex
            //set the texture mapping information
            pVertices->color = color;
            pVertices->tu = pSect[i].tu;
            pVertices->tv = 1.0f;
            CalcSectionVertex(pVertices, vertTo, pNormals[i], pSect[i],
                              bevelToUp, bevelToRight, scaleTo);
            ++pVertices;
        }
    }
    else
    {
        //fill in the missing triangle from the triangle strip if it is the
        //beginning of the loft triangle strip
        if(dwFlags & LOFT_ENDFROM)
        {
            //calculate the first vertex
            //set the texture mapping information
            pVertices->color = color;
            pVertices->tu = pSect[iSectSize-1].tu;
            pVertices->tv = 0.0f;
            CalcSectionVertex(pVertices, vertFrom, pNormals[iSectSize-1], pSect[iSectSize-1],
                              bevelFromUp, bevelFromRight, scaleFrom);
            ++pVertices;

            //calculate the second vertex
            //set the texture mapping information
            pVertices->color = color;
            pVertices->tu = pSect[0].tu;
            pVertices->tv = 1.0f;
            CalcSectionVertex(pVertices, vertTo, pNormals[0], pSect[0],
                              bevelToUp, bevelToRight, scaleTo);
            ++pVertices;
        }

        //draw the triangle strip
        for(int i = 0; i < iSectSize; ++i)
        {
            //calculate the first vertex
            //set the texture mapping information
            pVertices->color = color;
            pVertices->tu = pSect[i].tu;
            pVertices->tv = 0.0f;
            CalcSectionVertex(pVertices, vertFrom, pNormals[i], pSect[i],
                              bevelFromUp, bevelFromRight, scaleFrom);
            ++pVertices;

            //calculate the second vertex
            //set the texture mapping information
            pVertices->color = color;
            pVertices->tu = pSect[i].tu;
            pVertices->tv = 1.0f;
            CalcSectionVertex(pVertices, vertTo, pNormals[i], pSect[i],
                              bevelToUp, bevelToRight, scaleTo);
            ++pVertices;
        }


        //fill in the missing triangle from the triangle strip if it is the
        //end of the loft triangle strip
        if(dwFlags & LOFT_ENDTO)
        {
            //calculate the last vertex
            //set the texture mapping information
            pVertices->color = color;
            pVertices->tu = pSect[0].tu;
            pVertices->tv = 1.0f;
            CalcSectionVertex(pVertices, vertTo, pNormals[0], pSect[0],
                              bevelToUp, bevelToRight, scaleTo);
            ++pVertices;
        }
    }
    
    delete[] pNormals;
    return pVertices;
}

//draw a lofted section that is beveled based on a path and a section outline
//vertarray is an array that holds:
//      [vertfrom] [vertto] (bevelto) (scaleto)
//depending on the flags passed in
//pSect is the cross section to loft along the path
CUSTOMVERTEX* CLofter::LoftBeveledSectionIndexed(const D3DXVECTOR3* pVertArray,
                                                 const VECTOR2TEX*  pSect,
                                                 int                iSectSize,
                                                 CUSTOMVERTEX*      pVertices,
                                                 D3DCOLOR           color,
                                                 DWORD              dwFlags)
{
    return LoftBeveledSection(pVertArray, pSect, iSectSize, pVertices, color, dwFlags | LOFT_INDEXED | LOFT_NOBEVELFROM);
}

//used to calculate the indices for the vertices calculate in LoftBeveledSection
WORD* CLofter::CalcBeveledSectionIndices(WORD* pIndices,
                                         int iSectSize,
                                         int iNumSegments,
                                         DWORD dwFlags)
{
    _ASSERT(pIndices);

    //fill in the missing triangle at the beginning
    //adds one extra triangle to the beginning segment
    *pIndices = iSectSize - 1;
    ++pIndices;
    *pIndices = iSectSize;
    ++pIndices;

    for(int iCurSeg = 0; iCurSeg < iNumSegments; ++iCurSeg)
    {
        for(int iSectPos = 0; iSectPos < iSectSize; ++iSectPos)
        {
            *pIndices = iCurSeg * iSectSize + iSectPos;
            ++pIndices;

            *pIndices = iCurSeg * iSectSize + iSectSize + iSectPos;
            ++pIndices; 
        }
    }

    //fill in the missing triangle at the end
    *pIndices = (iNumSegments - 1) * iSectSize + iSectSize;
    ++pIndices;

    return pIndices;
}

//calculates the up and right vector based on the bevel between the two delta vectors
// and a right vector to the main vector in order to maintain width at the bevel
void CLofter::CalcBevelUpRight(D3DXVECTOR3*       pBevelUp,
                               D3DXVECTOR3*       pBevelRight,
                               const D3DXVECTOR3& delta,
                               const D3DXVECTOR3& deltaBevel,
                               const D3DXVECTOR3& right,
                               const D3DXVECTOR3& up)
{
    _ASSERT(pBevelRight && pBevelUp);

    //find the vector in between the two outgoing vectors for each point
    D3DXVECTOR3 vAverage = delta + deltaBevel;
    
    //normalize in the xy plane for delta and deltabevel;
    D3DXVECTOR2 deltaXY = D3DXVECTOR2(delta.x, delta.y);
    deltaXY /= D3DXVec2Length(&deltaXY);
    D3DXVECTOR2 deltaBevelXY = D3DXVECTOR2(deltaBevel.x, deltaBevel.y);
    deltaBevelXY /= D3DXVec2Length(&deltaBevelXY);
    *pBevelRight = D3DXVECTOR3(deltaXY.x + deltaBevelXY.x, deltaXY.y + deltaBevelXY.y, 0.0f);

    //resize vectors to be length one distance from the main shape
    //account for floating point inaccuracies
    FLOAT temp = (delta.x * pBevelRight->y - pBevelRight->x * delta.y);
    if(fabs(pBevelRight->x) > 1e-5f && fabs(temp) > 1e-5f)
    {
        FLOAT temp2 = pBevelRight->x * (delta.x * right.y  - right.x * delta.y) / temp;
        pBevelRight->y = pBevelRight->y / pBevelRight->x * temp2;
        pBevelRight->x = temp2;
    }
    else if(fabs(pBevelRight->y) > 1e-5f && fabs(delta.x) > 1e-5f)
    {
        //when x equals zero and the bevel is not on the same line
        pBevelRight->y = right.y - right.x * delta.y / delta.x;
    }
    else
    {
        *pBevelRight = right;
    }

    D3DXVECTOR3 vCross;
    D3DXVec3Cross(&vCross, pBevelRight, &vAverage);

    if(D3DXVec3LengthSq(&vCross) < 1e-5f)
    {
        *pBevelUp = up;
    }
    else
    {
        D3DXVec3Cross(pBevelUp, pBevelRight, &vCross);
        if(pBevelUp->z < 0.0f)
            *pBevelUp = -(*pBevelUp);

        //normalize the vector
        *pBevelUp /= D3DXVec3Length(pBevelUp);
    }
}

//calculates the vertices based on a given bevel 
void CLofter::CalcSectionVertex(CUSTOMVERTEX*      pVertex,
                                const D3DXVECTOR3& vSrcVert,
                                const D3DXVECTOR2& vNormal,
                                const VECTOR2TEX&  vtSectVert,
                                const D3DXVECTOR3& vBevelUp,
                                const D3DXVECTOR3& vBevelRight,
                                const D3DXVECTOR3& vScale)
{
    _ASSERT(pVertex);

    //calculate the first vertex
    pVertex->x = vSrcVert.x - vBevelRight.x * vtSectVert.x * vScale.x + vBevelUp.x * vtSectVert.y * vScale.y;
    pVertex->y = vSrcVert.y - vBevelRight.y * vtSectVert.x * vScale.x + vBevelUp.y * vtSectVert.y * vScale.y;
    pVertex->z = vSrcVert.z - vBevelRight.z * vtSectVert.x * vScale.x + vBevelUp.z * vtSectVert.y * vScale.y;

    //bias the normal based on the bevel
    pVertex->nx = -vBevelRight.x * vNormal.x;
    pVertex->ny = -vBevelRight.y * vNormal.x;
    pVertex->nz = vNormal.y - vBevelRight.z * vNormal.x;

    //normalize the normal
    FLOAT ilength = (FLOAT) sqrt(pVertex->nx*pVertex->nx + pVertex->ny*pVertex->ny + pVertex->nz*pVertex->nz);
    pVertex->nx /= ilength;
    pVertex->ny /= ilength;
    pVertex->nz /= ilength;
}