// NetworkManager.h: interface for the CNetworkManager class.
//
//////////////////////////////////////////////////////////////////////
#pragma once

#include "SnakeManager.h"
#include <dplay8.h>
#include <d3d8types.h>
#include <map>

using namespace std;

class CNetworkManager
{
public:
    CNetworkManager(const GUID& guidapp, CCourseLoader* pCourseLoader,
                    CSnakeManager* pSnakeMan, CItemManager* pItemMan,
                    UINT iMaxPlayers);
    ~CNetworkManager(void);

    //set the player info for the current player
    void SetLocalPlayerInfo(const TCHAR* pszName, D3DCOLOR snakecolor);

    //waits for connections
    HRESULT StartServer();

    //connects to server
    HRESULT StartClient(const WCHAR* pwszHostAddr);

    //sent by server to begin the game
    void BeginGame();

    //ends the game
    void EndGame();

    //sends the snake message to the network
    void SendSnakeMessage(DWORD dwAction);

    //sends the snake synchonization message
    void SendSnakeSynch(BOOL bIsDeadMessage = FALSE);

    //sends the snake length
    void SendSnakeState();

    //sends the initial positioning information for all the snakes
    void SendSnakePositions();

    //sends the initial positioning information for all the items
    void SendItemPositions(DPNID dpnid);

    //sends the selected course
    void SendSelectedCourse(DPNID dpnid);

    //cancels the client attempt to connect to host
    void CancelClient();

    //waits for the start signal BeginGame() the server to be called
    void WaitForStartSignal();

    //returns whether a connection was established or not
    BOOL GetIsConnected(); 

    //returns whether or not the local machine is a game server
    BOOL GetIsServer();

    //returns whether the client is ready to start
    BOOL GetIsReadyToStart();

    //returns whether the game has started yet
    BOOL GetGameStarted();

    //returns the local IP address
    WCHAR* GetLocalIPAddress();

    //ends all network connections
    void EndConnection();

    struct PlayerInfo
    {
        PlayerInfo() : m_SnakeColor(D3DCOLOR_RGBA(0,255,0,255)),
                       m_pSnake(NULL), m_bReadyToStart(FALSE)
        {
            m_pszName[0] = '\0';
        }

        TCHAR    m_pszName[256];
        D3DCOLOR m_SnakeColor;
        CSnake*  m_pSnake;
        BOOL     m_bReadyToStart;
    };

    //returns the information of the winning snake or NULL if none
    PlayerInfo* GetWinningSnakeInfo();

    //creates all of the CSnake objects and sets their color
    void RegenerateSnakeObjects();

    static map<DPNID, PlayerInfo>  m_Players;
    static DPNID                   m_LocalPlayerID;
private:

    //the local player description
    PlayerInfo                     m_LocalPlayer;
    
    //DirectPlay callback handler for the server
    static HRESULT WINAPI ServerCallback(PVOID pvUserContext, DWORD dwMessageType, PVOID pMessage);

    //DirectPlay callback  handler for the client
    static HRESULT WINAPI ClientCallback(PVOID pvUserContext, DWORD dwMessageType, PVOID pMessage);
    
    //application specific messages
    enum APP_MESSAGES
    {
        MSG_PLAYERINFO  = 0,    //sends player name and color
        MSG_LOCALID     = 1,    //sends local DPNID back to client
        MSG_SNAKEACTION = 2,    //sends a snake action(turn, jump, dive)
        MSG_SNAKEPOS    = 3,    //sends the snake positions
        MSG_ITEMPOS     = 4,    //sends the item positions
        MSG_BEGINGAME   = 5,    //sends the begin game to clients / ready for play to server
        MSG_SNAKESTATE  = 6,    //sends snake state(length, speed)
        MSG_SYNCSNAKE   = 7,    //sends all of the snake segments
        MSG_SETCOURSE   = 8,    //sends the filename of course file
        MSG_SNAKEDEAD   = 9,    //sends the local snake is dead message
        MSG_SNAKELEAVE  = 10    //sends a snake leave message
    };

    //Server Application Specific Message Handler
    HRESULT ServerMessageProc(PDPNMSG_RECEIVE pMsg);

    //Client Application Specific Message Handler
    HRESULT ClientMessageProc(PDPNMSG_RECEIVE pMsg);

    //structure to keep track of whether all info was recieved yet
    struct RecieveInfo
    {
        RecieveInfo()
        {
            Reset();
        }

        inline BOOL AllRecieved()
        {
            return bCourseRecieved && bSnakePosRecieved && bItemPosRecieved;
        }

        inline void Reset()
        {
            bCourseRecieved   = FALSE;
            bSnakePosRecieved = FALSE;
            bItemPosRecieved  = FALSE;
        }

        BOOL bCourseRecieved;
        BOOL bSnakePosRecieved;
        BOOL bItemPosRecieved;
    } m_RecieveInfo;

    UINT                              m_iMaxPlayers;
    GUID                              m_GUIDApp;
    CCourseLoader*                    m_pCourseLoader;
    CSnakeManager*                    m_pSnakeManager;
    CItemManager*                     m_pItemManager;
    static IDirectPlay8Address*       m_pDeviceAddress;
    static IDirectPlay8Server*        m_pServer;
    static IDirectPlay8Client*        m_pClient;
    static DPNHANDLE                  m_hEnumAsyncOp;
    BOOL                              m_bIsConnected;
    BOOL                              m_bGameStarted;
    BOOL                              m_bHostConnectLost;
    WCHAR                             m_pwszLocalAddr[256];
};