// NetworkMenu.cpp: implementation of the CNetworkMenu class.
//
//////////////////////////////////////////////////////////////////////

#include "NetworkMenu.h"
#include <crtdbg.h>     //for _ASSERT

CAbstractMenu::MenuItem CNetworkMenu::m_pNetworkMenuItems[] =
{
    {"Host Game",       CAbstractMenu::MENUID_HOST},
	{"Join Game",       CAbstractMenu::MENUID_CLIENT},
    {"Main Menu",       CAbstractMenu::MENUID_MAINMENU}
};

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CNetworkMenu::CNetworkMenu() : m_fRotAngle(3.1415f / 2),
                           m_fTexOffset(0.0f),
                           m_iCurSel(0)
{
    m_pFont      = new CD3DFont( _T("Arial"), 20, D3DFONT_BOLD );
    m_pTitleMesh = new CD3DMesh();
}

CNetworkMenu::~CNetworkMenu()
{
    DeleteDeviceObjects();
}

int CNetworkMenu::OnKeyDown(WPARAM wParam)
{
    switch(wParam)
    {
    case VK_RETURN:
        return m_pNetworkMenuItems[m_iCurSel].m_iMenuID;
        break;
    case VK_DOWN:
        ++m_iCurSel;
        m_iCurSel %= sizeof(m_pNetworkMenuItems) / sizeof(CAbstractMenu::MenuItem);
        break;
    case VK_UP:
        --m_iCurSel;
        if(m_iCurSel < 0)
            m_iCurSel = sizeof(m_pNetworkMenuItems) / sizeof(CAbstractMenu::MenuItem) - 1;
        break;
    }
    return CAbstractMenu::MENUID_NOTHING;
}

int CNetworkMenu::FrameMove()
{
    _ASSERT(m_pd3dDevice);
    
    if(m_fRotAngle > 0.0f)
        m_fRotAngle -= 0.01f;
    else
        m_fRotAngle = 0.0f;
    

    m_fTexOffset += 0.003f;
    if(m_fTexOffset > 1.0f)
        m_fTexOffset -= 1.0f;

    return CAbstractMenu::MENUID_NOTHING;
}

//render the title menu
void CNetworkMenu::RenderMenu(int iWidth, int iHeight)
{
    _ASSERT(m_pTitleMesh && m_pd3dDevice);
    m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, 0x00000000, 1.0f, 0L );
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2 );
    
    //set up the camera view position
    D3DXMATRIX matView;
    D3DXVECTOR3 vLookatPt   = D3DXVECTOR3( 0.0f, -15.0f, 0.0f );
    D3DXVECTOR3 vUpVec      = D3DXVECTOR3( 0.0f, 1.0f, 0.0f );
    D3DXVECTOR3 vEyePt      = D3DXVECTOR3( 0.0f, -15.0f, 100.0f);
    D3DXMatrixLookAtLH( &matView, &vEyePt, &vLookatPt, &vUpVec );
    m_pd3dDevice->SetTransform( D3DTS_VIEW, &matView );

    D3DXMATRIX matWorld;
    D3DXMatrixRotationY(&matWorld, 3.1415f);
    m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

    D3DXMatrixRotationX(&matWorld, -m_fRotAngle);
    m_pd3dDevice->MultiplyTransform( D3DTS_WORLD, &matWorld );

    //set to translate texture
    // coordinates in the U and V directions.
    //      1   0  du  0
    //      0   1  dv  0
    //      0   0  1   0
    //      0   0  0   1
    D3DXMATRIX matTex;
    D3DXMatrixIdentity(&matTex); // declared in d3dutil.h
    matTex._31 = -m_fTexOffset;
    matTex._32 = m_fTexOffset;
    m_pd3dDevice->SetTransform(D3DTS_TEXTURE0, &matTex);
    
    //render the title mesh
    m_pTitleMesh->Render( m_pd3dDevice );
    RenderText(iWidth, iHeight);

    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE );
}


HRESULT CNetworkMenu::RenderText(int iWidth, int iHeight)
{
    _ASSERT(m_pFont);
    
    D3DCOLOR fontColor        = D3DCOLOR_ARGB(255,128,128,128);
    D3DCOLOR fontSelectColor  = D3DCOLOR_ARGB(255,128,0,255);

    // Output display stats
    FLOAT fNextLine = 0.0f; 
    SIZE txtsize;

    //place brackets around the selected item
    char pszTemp[256];
    pszTemp[0] = '[';
    strcpy(pszTemp + 1, m_pNetworkMenuItems[m_iCurSel].m_pszMenuStr);
    strcat(pszTemp, "]");

    int iNumItems = sizeof(m_pNetworkMenuItems) / sizeof(CAbstractMenu::MenuItem);
    for(int i = 0; i < iNumItems; ++i)
    {
        if(i == m_iCurSel)
        {
            m_pFont->GetTextExtent( pszTemp, &txtsize );
            m_pFont->DrawText( (iWidth - txtsize.cx) / 2.0f,
                               3.0f * (iHeight - txtsize.cy) / 4.0f - txtsize.cy / 2.0f + fNextLine,
                               fontSelectColor,
                               pszTemp );
        }
        else
        {
            m_pFont->GetTextExtent( m_pNetworkMenuItems[i].m_pszMenuStr, &txtsize );
            m_pFont->DrawText( (iWidth - txtsize.cx) / 2.0f,
                               3.0f * (iHeight - txtsize.cy) / 4.0f - txtsize.cy / 2.0f + fNextLine,
                               fontColor,
                               m_pNetworkMenuItems[i].m_pszMenuStr );
        }
        fNextLine += txtsize.cy;
    }

    return S_OK;
}


HRESULT CNetworkMenu::InitDeviceObjects( LPDIRECT3DDEVICE8 pd3dDevice )
{
    _ASSERT(m_pTitleMesh && m_pFont && pd3dDevice);

    // Keep a local copy of the device
    m_pd3dDevice = pd3dDevice;

    if( FAILED( m_pTitleMesh->Create( m_pd3dDevice, _T("Core\\title.x") ) ) )
        return E_FAIL;

    // Init the font
    m_pFont->InitDeviceObjects( m_pd3dDevice );

    return S_OK;
}

HRESULT CNetworkMenu::RestoreDeviceObjects()
{
    _ASSERT(m_pd3dDevice && m_pTitleMesh && m_pFont);
    m_pTitleMesh->RestoreDeviceObjects( m_pd3dDevice );
    m_pFont->RestoreDeviceObjects();
    return S_OK;
}

HRESULT CNetworkMenu::InvalidateDeviceObjects()
{
    _ASSERT(m_pTitleMesh && m_pFont);
    m_pTitleMesh->InvalidateDeviceObjects();
    m_pFont->InvalidateDeviceObjects();
    return S_OK;
}

HRESULT CNetworkMenu::DeleteDeviceObjects()
{
    if(m_pTitleMesh)
        m_pTitleMesh->Destroy();
    if(m_pFont)
        m_pFont->DeleteDeviceObjects();
    m_pd3dDevice = NULL;

    SAFE_DELETE( m_pTitleMesh );
    SAFE_DELETE( m_pFont );
    return S_OK;
}