// Snake.cpp: implementation of the CSnake class.
//
//////////////////////////////////////////////////////////////////////

#include "Snake.h"
#include "DXUtil.h"
#include "d3dutil.h"
#include "GameMath.h"
#include <crtdbg.h>   //for _ASSERT

#define SNAKE_TEX_REPEAT    10.0f
#define SNAKE_TAIL_TAPER    5.0f
#define SNAKE_VB_SIZE       32
#define SNAKE_ROTATE_AMT    3.1415f / 8.0f

VECTOR2TEX CSnake::m_pSnakeSect[] =
{
    //  xpos        ypos        tu
    VECTOR2TEX(  0.0f,        1.0f,       0.0f    ),
    //VECTOR2TEX(  0.7071f,     0.7071f,    0.25f   ),
    VECTOR2TEX(  1.0f,        0.0f,       0.50f   ),
    //VECTOR2TEX(  0.7071f,     -0.7071f,   0.75f   ),
    VECTOR2TEX(  0.0f,        -1.0f,      1.00f   ),
    //VECTOR2TEX(  -0.7071f,    -0.7071f,   0.75f   ),
    VECTOR2TEX( -1.0f,        0.0f,       0.50f   ),
    //VECTOR2TEX( -0.7071f,     0.7071f,    0.25f   ),
};

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CSnake::CSnake()
{
    m_pd3dDevice           = NULL;
    m_pIndexBuf            = NULL;
    m_pVB                  = NULL;
    m_pTexture             = NULL;
    m_Color                = D3DCOLOR_RGBA(0, 255, 0, 255);
    m_pSnakeHead           = new CD3DMesh();
}

CSnake::~CSnake()
{
    InvalidateDeviceObjects();
    DeleteDeviceObjects();

    SAFE_DELETE( m_pSnakeHead );
}

//reset the snake's position, size, speed, movedir, and path 
void CSnake::ResetPosition(const D3DXVECTOR3& pos, const D3DXVECTOR3& movedir)
{
    m_PathVect.resize(0);

    //push back two points because the last point is always modified
    m_PathVect.push_back(PATH_VERTEX(pos, 0.0f));
    m_PathVect.push_back(PATH_VERTEX(pos + movedir));
    m_vMoveDir = movedir;
}

//resets the state of the snake
void CSnake::ResetState()
{
    m_fCurLength           = 1.0f;
    m_fTargetLength        = 15.0f;
    m_fScaleRadius         = 1.0f;
    
    m_dwActionSeq          = ACTION_NONE;
    m_iActionSeqFrame      = 0;

    m_iScore               = 0;
    m_bIsDead              = FALSE;
    m_fSpeed               = 0.2f;
}

HRESULT CSnake::InitDeviceObjects( LPDIRECT3DDEVICE8 pd3dDevice )
{
    // Keep a local copy of the device
    m_pd3dDevice = pd3dDevice;

    // Load the snake head
    if( FAILED( m_pSnakeHead->Create( m_pd3dDevice, _T("Core\\snakehead.x") ) ) )
        return E_FAIL;

    // Create a new texture for snake body
    if( FAILED( D3DXCreateTextureFromFile( m_pd3dDevice, "Core\\snake.jpg", &m_pTexture ) ) )
        return E_FAIL;

    return S_OK;
}

HRESULT CSnake::RestoreDeviceObjects()
{
    m_pSnakeHead->RestoreDeviceObjects( m_pd3dDevice );

    //create the vertex buffer and indices for the snake
    int iVertPerSect = sizeof(m_pSnakeSect) / sizeof(VECTOR2TEX);
    int iVbSize      = sizeof(CUSTOMVERTEX) * iVertPerSect * SNAKE_VB_SIZE;    //store SNAKE_VB_SIZE segments
    
    if( FAILED( m_pd3dDevice->CreateVertexBuffer( iVbSize, D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, &m_pVB ) ) )
        return E_FAIL;

    if( FAILED( m_pd3dDevice->CreateIndexBuffer( 2 * iVbSize, D3DUSAGE_WRITEONLY, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &m_pIndexBuf ) ) )
        return E_FAIL;

    //fill the index buffer
    WORD* pIndices;
    if( FAILED( m_pIndexBuf->Lock( 0, 0, (BYTE**)&pIndices, 0  ) ) )
        return E_FAIL;
    CLofter::CalcBeveledSectionIndices(pIndices, iVertPerSect, SNAKE_VB_SIZE);
    m_pIndexBuf->Unlock();

    return S_OK;
}

HRESULT CSnake::InvalidateDeviceObjects()
{
    m_pSnakeHead->InvalidateDeviceObjects();
    SAFE_RELEASE( m_pIndexBuf );
    SAFE_RELEASE( m_pVB );
    return S_OK;
}

HRESULT CSnake::DeleteDeviceObjects()
{
    m_pSnakeHead->Destroy();
    SAFE_RELEASE( m_pIndexBuf );
    SAFE_RELEASE( m_pVB );
    SAFE_RELEASE( m_pTexture );
    m_pd3dDevice = NULL;
    return S_OK;
}

//does a specified action based on the flag
void CSnake::DoAction(DWORD dwAction, CD3DMesh* pTerrain)
{
    //only do actions when move is called
    if(dwAction == ACTION_MOVE)
    {
        if(m_dwActionSeq & ACTION_DEATH)
        {
            DeathSequence();
            return;
        }

        if(m_dwActionSeq & ACTION_STOP)
        {
            //unset the stop flag
            m_dwActionSeq &= ~ACTION_STOP;
            return;
        }

        if(m_dwActionSeq & ACTION_TURNLEFT ||
           m_dwActionSeq & ACTION_TURNRIGHT)
        {
            Turn();

            //if there are other flags then just turn off the turning flags
            if(m_dwActionSeq & ~ACTION_TURNLEFT & ~ACTION_TURNRIGHT)
            {
                //unset the turning flags
                m_dwActionSeq &= ~ACTION_TURNLEFT;
                m_dwActionSeq &= ~ACTION_TURNRIGHT;
            }
        }

        //find the height of the terrain in the given direction
        FLOAT fTerrainDelta = -100.0f;
        if(pTerrain)
        {
            FLOAT       fDist;
            BOOL        bHit;
            D3DXVECTOR3 vRayPos  = *GetLeadPoint() + GetDirection();
            vRayPos.z           += 4.0f;     //check for terrain intersect a little above head
            D3DXVECTOR3 vRay     = D3DXVECTOR3(0.0f, 0.0f, -1.0f);
            if(SUCCEEDED( D3DXIntersect(pTerrain->GetSysMemMesh(),
                                        &vRayPos, &vRay, &bHit,
                                        NULL, NULL, NULL, &fDist, NULL, NULL) ) && bHit)
            {
                fTerrainDelta = 4.0f - fDist;
            }
        }

        if(m_dwActionSeq & ACTION_JUMP)
        {
            Jump(fTerrainDelta);
            return;
        }

        if(m_dwActionSeq & ACTION_DIVE)
        {
            Dive(fTerrainDelta);
            return;
        }
        
        FLOAT fMoveAmt = m_fSpeed;
        //conform to terrain if it is out of the height range and
        //the length of the segment is larger than 1 or it is turning
        if(GetLeadSegmentLenSq() > 0.0f &&
           (m_dwActionSeq & ACTION_TURNLEFT || m_dwActionSeq & ACTION_TURNRIGHT || fabs(fTerrainDelta + 1.0f) > 0.1f))
        {
            D3DXVECTOR3 vLeadDelta = GetDirection();

            //place it on the terrain if it is close enough
            if(fabs(fTerrainDelta) < 0.9f)
            {
                vLeadDelta.z += fTerrainDelta + 1.0f;
            }
            else
            {
                //if the snake is falling or moving up then move it by a constant speed
                //if it is not already within the max angle
                if(vLeadDelta.z < 0.9 && fTerrainDelta > 0.0f)
                    vLeadDelta.z +=  0.1f;
                else if(vLeadDelta.z > -0.9 && fTerrainDelta < 0.0f)
                    vLeadDelta.z -= 0.1f;
                fMoveAmt = 0.5f;
            }
           
            vLeadDelta /= D3DXVec3Length(&vLeadDelta);
            m_vMoveDir = vLeadDelta;

            //don't generate a new vertex if in the eating sequence
            if(!(m_dwActionSeq & ACTION_EAT))
                m_PathVect.push_back(m_PathVect.back());
        }

        //unset the turning flags
        m_dwActionSeq &= ~ACTION_TURNLEFT;
        m_dwActionSeq &= ~ACTION_TURNRIGHT;

        if(m_dwActionSeq & ACTION_EAT)
            Eat();

        if(!m_dwActionSeq)
            Move(fMoveAmt);
    }
    else if(m_dwActionSeq != ACTION_DEATH)
    {
        //eating and dying override current action,
        //otherwise, actions are exclusive
        //turning can be tacked onto any of the actions
        if(m_dwActionSeq == ACTION_NONE  ||
           dwAction      == ACTION_EAT   ||
           dwAction      == ACTION_DEATH)
        {
            m_dwActionSeq     = dwAction;
            m_iActionSeqFrame = 0;
        }
        else if(dwAction == ACTION_TURNRIGHT ||
                dwAction == ACTION_TURNLEFT ||
                dwAction == ACTION_STOP)
        {
            m_dwActionSeq |= dwAction;
        }
    }
}

//move the snake by a multiplier amount
void CSnake::Move(FLOAT speed)
{
    //move the front of the snake forward
    m_PathVect.back().vPos += m_vMoveDir * speed;

    //do not move the tail if the current length is
    //shorter than the target length
    if(m_fCurLength < m_fTargetLength)
    {
        m_fCurLength += speed;
    }
    else //move the tail of the snake forward
    {
        MoveSnakeTail(speed);
        
        //if snake is too long then shorten the snake
        if(m_fCurLength > m_fTargetLength)
        {
            MoveSnakeTail(0.1f);
            m_fCurLength -= 0.1f;
        }
    }    
}

void CSnake::MoveSnakeTail(FLOAT moveamt)
{
    D3DXVECTOR3 vDiffVect = m_PathVect[1].vPos - m_PathVect[0].vPos;
    FLOAT fDiffLen = D3DXVec3Length(&vDiffVect);

    //if the move amount is more than the length of the
    //segment delete the segment
    while(fDiffLen < moveamt && m_PathVect.size())
    {
        m_PathVect.erase(m_PathVect.begin());
        moveamt -= fDiffLen;

        vDiffVect = m_PathVect[1].vPos - m_PathVect[0].vPos;
        fDiffLen = D3DXVec3Length(&vDiffVect);
    }

    if(!m_PathVect.size())
        return;

    //move the tail
    D3DXVECTOR3 vChangeVect;
    D3DXVec3Normalize(&vChangeVect, &vDiffVect);
    m_PathVect[0].vPos += vChangeVect * moveamt;
}

//signal that the snake has just eaten something
void CSnake::Eat()
{
    //create first vertex
    FLOAT m_fSpeedLeft = m_fSpeed;
    if(m_iActionSeqFrame == 0)
    {
        m_PathVect.push_back(m_PathVect.back());
        m_PathVect.back().fScale = 1.75f;
        ++m_iActionSeqFrame;
    }

    //move the front of the bulge
    if(m_iActionSeqFrame == 1)
    {
        FLOAT fDistLeft = (float)(1.5f - sqrt(GetLeadSegmentLenSq()));
        if(m_fSpeedLeft < fDistLeft)
        {
            Move(m_fSpeedLeft);
            fDistLeft -= m_fSpeedLeft;
            m_fSpeedLeft = 0.0f;
        }
        else
        {
            Move(fDistLeft);
            m_fSpeedLeft -= fDistLeft;
            fDistLeft = 0.0f;
        }

        if(fDistLeft <= 0.0f)
            ++m_iActionSeqFrame;
    }
    
    //create second vertex
    if(m_iActionSeqFrame == 2)
    {
        m_PathVect.push_back(m_PathVect.back());
        m_PathVect.back().fScale = 1.0f;
        ++m_iActionSeqFrame;
    }

    //move the end of the bulge
    if(m_iActionSeqFrame == 3)
    {
        FLOAT fDistLeft = (float)(1.5f - sqrt(GetLeadSegmentLenSq()));
        if(m_fSpeedLeft < fDistLeft)
        {
            Move(m_fSpeedLeft);
            fDistLeft -= m_fSpeedLeft;
            m_fSpeedLeft = 0.0f;
        }
        else
        {
            Move(fDistLeft);
            m_fSpeedLeft -= fDistLeft;
            fDistLeft = 0.0f;
        }

        if(fDistLeft <= 0.0f)
            ++m_iActionSeqFrame;
    }

    if(m_iActionSeqFrame == 4)
    {
        m_PathVect.push_back(m_PathVect.back());
        Move(m_fSpeedLeft);

        m_dwActionSeq     &= ~ACTION_EAT;
        m_iActionSeqFrame = 0;
    }
}

//turns the snake
void CSnake::Turn()
{
    D3DXVECTOR3 turndir;
    if(m_dwActionSeq & ACTION_TURNLEFT)
    {
        turndir.x = (float)(cos(SNAKE_ROTATE_AMT) * m_vMoveDir.x - sin(SNAKE_ROTATE_AMT) * m_vMoveDir.y);
        turndir.y = (float)(sin(SNAKE_ROTATE_AMT) * m_vMoveDir.x + cos(SNAKE_ROTATE_AMT) * m_vMoveDir.y);
    }
    else if(m_dwActionSeq & ACTION_TURNRIGHT)
    {
        turndir.x = (float)(cos(-SNAKE_ROTATE_AMT) * m_vMoveDir.x - sin(-SNAKE_ROTATE_AMT) * m_vMoveDir.y);
        turndir.y = (float)(sin(-SNAKE_ROTATE_AMT) * m_vMoveDir.x + cos(-SNAKE_ROTATE_AMT) * m_vMoveDir.y);
    }
    turndir.z = m_vMoveDir.z;

    //m_vMoveDir = turndir;
    //m_PathVect.push_back(m_PathVect.back());
    SetDirection(turndir);
}

//signal that the snake should jump
void CSnake::Jump(FLOAT fTerrainDelta)
{
    D3DXVECTOR3 m_vCurDir = m_vMoveDir;
    if(m_iActionSeqFrame == 0)
    {
        m_vCurDir.z = 1.0f;
        m_vCurDir /= D3DXVec3Length(&m_vCurDir);
        SetDirection(m_vCurDir);
        Move(0.5f);
        ++m_iActionSeqFrame;
    }
    else if(m_iActionSeqFrame == 1)
    {
        //set the initial velocity
        if(fTerrainDelta >= -2.0f)
        {
            Move(0.5f);
        }
        else
        {
            m_vCurDir.z = 4.0f;
            m_vCurDir /= D3DXVec3Length(&m_vCurDir);
            SetDirection(m_vCurDir);
            Move(0.75f);
            ++m_iActionSeqFrame;
        }
    }
    else if(m_iActionSeqFrame % 8 == 2)
    {
        //set the gravity
        m_vCurDir.z -= 0.6f;
        if(m_vCurDir.z < -1.0f)
            m_vCurDir.z = -1.0f;
        if(fTerrainDelta >= -3.0f)
            m_vCurDir.z = -0.20f;
        m_vCurDir /= D3DXVec3Length(&m_vCurDir);
        SetDirection(m_vCurDir);
        ++m_iActionSeqFrame;
        Move(0.5f);
    }
    else
    {
        //move in between the frames
        Move(0.5f);
        ++m_iActionSeqFrame;
    }

    //conform to terrain
    if(fTerrainDelta > -0.5f)
    {
        m_dwActionSeq     &= ~ACTION_JUMP;
        m_iActionSeqFrame = 0;
    }
}

//signal that the snake should dive
void CSnake::Dive(FLOAT fTerrainDelta)
{
    D3DXVECTOR3 m_vCurDir = m_vMoveDir;
    if(m_iActionSeqFrame == 0)
    {
        m_vCurDir.z = -0.5f;
        m_vCurDir /= D3DXVec3Length(&m_vCurDir);
        SetDirection(m_vCurDir);
        Move(0.2f);
        ++m_iActionSeqFrame;
    }
    else if(m_iActionSeqFrame == 1)
    {
        //set the initial velocity
        if(fTerrainDelta > 3.0f)
        {
            m_vCurDir.z = 0.5f;
            m_vCurDir /= D3DXVec3Length(&m_vCurDir);
            SetDirection(m_vCurDir);
            Move(m_fSpeed);
            ++m_iActionSeqFrame;
        }
        else
        {
            Move(m_fSpeed);
        }
    }
    else
    {
        //move in between the frames
        Move(m_fSpeed);
		++m_iActionSeqFrame;
    }

    //conform to terrain
    if(fTerrainDelta < -1.5f)
    {
        m_dwActionSeq     &= ~ACTION_DIVE;
        m_iActionSeqFrame = 0;
    }
}

//perform the snake's way too elaborate death sequence
void CSnake::DeathSequence()
{
    m_bIsDead = TRUE;
    
    //stop after the tail has reached 100
    /*if(m_PathVect[0].vPos.z > 100.0f)
    {
        m_bIsDead = TRUE;
        return;
    }

    if(GetLeadSegmentLenSq() > 10.0f)
    {
        FLOAT temp = m_vMoveDir.x;
        m_vMoveDir.x = (float)(cos(3.1415f / 7.0f) * temp - sin(3.1415f / 7.0f) * m_vMoveDir.y);
        m_vMoveDir.y = (float)(sin(3.1415f / 7.0f) * temp + cos(3.1415f / 7.0f) * m_vMoveDir.y);
        m_PathVect.push_back(m_PathVect.back());
    }

    m_vMoveDir.z = 0.5f;
    m_vMoveDir /= D3DXVec3Length(&m_vMoveDir);
    m_fSpeed = 0.5f;
    Move(m_fSpeed);*/
}

//get the speed of the snake
FLOAT CSnake::GetSpeed()
{
    return m_fSpeed;
}

//get the length squared of the leading segment
FLOAT CSnake::GetLeadSegmentLenSq()
{
    D3DXVECTOR3 vDiff = m_PathVect[m_PathVect.size() - 1].vPos - m_PathVect[m_PathVect.size() - 2].vPos;
    return D3DXVec3LengthSq(&vDiff);
}

//gets the two points that define the leading segment
void CSnake::GetLeadSegment(D3DXVECTOR3* pFrom, D3DXVECTOR3* pTo)
{
    if(pTo)
        *pTo      = m_PathVect[m_PathVect.size() - 1].vPos;
    if(pFrom)
        *pFrom    = m_PathVect[m_PathVect.size() - 2].vPos;
}

//gets the leading point of the path
D3DXVECTOR3* CSnake::GetLeadPoint()
{
    return &m_PathVect.back().vPos;
}

//get the direction of the snake
D3DXVECTOR3 CSnake::GetDirection()
{
    return m_vMoveDir;
}

//returns whether the segment defined by vFrom and vTo intersects the snake
//and the from and to points of the section it intersected
//if this is a self intersection then don't check the last segment
BOOL CSnake::GetCollision(const D3DXVECTOR3& vFrom, const D3DXVECTOR3& vTo,
                          const CSnake* pSnake,
                          D3DXVECTOR3* pvSectFrom, D3DXVECTOR3* pvSectTo)
{
    FLOAT       fZdelta;
    FLOAT       fPercentIntersect;  //the percent between vFrom and vTo of the intersection from 0.0f to 1.0f
    D3DXVECTOR3 vIntersect;

    //should take more than 3 sections to wrap around and hit itself
    int endpos = m_PathVect.size() - 1 - ((pSnake == this) ? 3 : 0);
    for(int pathpos = 0; pathpos < endpos; ++pathpos)
    {
        if(CGameMath::GetLineIntersect(vFrom, vTo,
                                       m_PathVect[pathpos].vPos,
                                       m_PathVect[pathpos + 1].vPos, NULL, &fZdelta) && fabs(fZdelta) < m_fScaleRadius)
        {
            if(pvSectFrom)
                *pvSectFrom = m_PathVect[pathpos].vPos;
            if(pvSectTo)
                *pvSectTo = m_PathVect[pathpos + 1].vPos;
            return TRUE;
        }
        else if(CGameMath::GetPointPerp(m_PathVect[pathpos].vPos, m_PathVect[pathpos + 1].vPos, vTo,
                                        &vIntersect, &fPercentIntersect))
        {
            FLOAT fSectionWidth = (1.0f - fPercentIntersect) * m_PathVect[pathpos].fScale * m_fScaleRadius +
                                  fPercentIntersect * m_PathVect[pathpos + 1].fScale * m_fScaleRadius;

            vIntersect -= vTo;
            if(D3DXVec3LengthSq(&vIntersect) < 2.0f * fSectionWidth * fSectionWidth)
            {
                if(pvSectFrom)
                    *pvSectFrom = m_PathVect[pathpos].vPos;
                if(pvSectTo)
                    *pvSectTo = m_PathVect[pathpos + 1].vPos;
                return TRUE;
            }
        }
    }

    return FALSE;
}

//returns the current state of the snake
DWORD CSnake::GetCurrentState()
{
    return m_dwActionSeq;
}

//returns whether the snake is dead or not
BOOL CSnake::GetIsDead()
{
    return (m_bIsDead || GetLeadPoint()->z < -50.0f);
}

//change the direction of the snake
void CSnake::SetDirection(const D3DXVECTOR3& movedir)
{
    if(m_vMoveDir != movedir && !(m_dwActionSeq & ACTION_DEATH))
    {
        m_vMoveDir = movedir;
        if(!(m_dwActionSeq & ACTION_EAT))
        {
            m_PathVect.push_back(m_PathVect.back());
            Move(0.01f);
        }
    }
}

//sets the color of the snake
void CSnake::SetColor(D3DCOLOR color)
{
    m_Color = color;
}

//set the speed of the snake
void CSnake::SetSpeed(FLOAT speed)
{
    m_fSpeed = speed;
}

//increase/decrease the length of the snake
void CSnake::ModifyLength(FLOAT incamt)
{
    m_fTargetLength += incamt;

    //make sure there is a minimum length
    if(m_fTargetLength < 5.0f)
        m_fTargetLength = 5.0f;
}

//draw the snake on the given Direct3DDevice
void CSnake::Render()
{
    _ASSERT(m_pVB && m_pd3dDevice);

    m_pd3dDevice->SetTexture( 0, m_pTexture);
    m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
    m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
    m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);

    m_pd3dDevice->SetVertexShader( D3DFVF_CUSTOMVERTEX );
    m_pd3dDevice->SetIndices(m_pIndexBuf, 0);
    m_pd3dDevice->SetStreamSource( 0, m_pVB, sizeof(CUSTOMVERTEX) );

    //two vertices per vertex
    UINT  iSectSize      = sizeof(m_pSnakeSect) / sizeof(VECTOR2TEX);
    FLOAT fTexOffset     = m_fCurLength / SNAKE_TEX_REPEAT;

    //used to taper the tail
    FLOAT         fDistRendered   = 0.0f;
    CUSTOMVERTEX* pLastVertexSect = new CUSTOMVERTEX[iSectSize];

    //draw SNAKE_VB_SIZE sections at a time until the snake is completely drawn
    UINT   iNumSegsDrawn  = 0;
    while(iNumSegsDrawn < m_PathVect.size() - 1)
    {
        int iMaxSegs = (m_PathVect.size() - 1 - iNumSegsDrawn > SNAKE_VB_SIZE - 2) ? SNAKE_VB_SIZE - 2 : m_PathVect.size() - iNumSegsDrawn - 1;

        //lock two extra sections just in case it gets split to create a better tail taper
        //or extra vertices are needed for the beginning and end lofted vertices
        CUSTOMVERTEX* pVertices;
        if( FAILED( m_pVB->Lock( 0, 0, (BYTE**)&pVertices, D3DLOCK_DISCARD  ) ) )
            return;
        CUSTOMVERTEX* pVerticesBegin = pVertices;
        
        //draw the snake's path
        D3DXVECTOR3 loftverts[5];
        D3DXVECTOR3 vDiff;
        FLOAT fLength;
        DWORD dwFlags = 0;

        //counter used to identify which part of a split segment is being drawn when
        //tapering the tail
        UINT  iCurSegmentSplit = 0;

        //stores the difference from the iMaxSegs: invalid loft = -1, split segment = +1
        int   iSegDifference = 0;

        //copy the last vertex section from the previous vertex buffer if this is a
        //continuation of the previous vertex buffer
        if(iNumSegsDrawn)
        {
            memcpy(pVertices, pLastVertexSect, sizeof(CUSTOMVERTEX) * iSectSize);
            pVertices += iSectSize;
        }

        for(UINT pathpos = iNumSegsDrawn; pathpos < iMaxSegs + iNumSegsDrawn; ++pathpos)
        {
            CUSTOMVERTEX* pVerticesSegBegin = pVertices;

            //path vertices
            loftverts[0] = m_PathVect[pathpos].vPos;
            loftverts[1] = m_PathVect[pathpos + 1].vPos;

            vDiff = loftverts[1] - loftverts[0];
            fLength = D3DXVec3Length(&vDiff);
            if(fLength < 1e-5f)
            {
                --iSegDifference;
                 continue;
            }

            //bevels
            loftverts[2] = m_PathVect[(pathpos >= m_PathVect.size() - 2) ? pathpos + 1 : pathpos + 2].vPos;

            //flags
            dwFlags = (fDistRendered == 0) ? CLofter::LOFT_ENDFROM : CLofter::LOFT_DEFAULT;
            if(pathpos == iNumSegsDrawn + iMaxSegs - 1)
                dwFlags |= CLofter::LOFT_ENDTO;
            
            //taper the tail
            if(fDistRendered < SNAKE_TAIL_TAPER)
            {
                if(fLength < SNAKE_TAIL_TAPER - fDistRendered + 0.01f)
                {
                    loftverts[3].x = loftverts[3].y = loftverts[3].z = m_PathVect[pathpos].fScale * fDistRendered / SNAKE_TAIL_TAPER;
                    FLOAT fToScale = (fDistRendered + fLength) / SNAKE_TAIL_TAPER;
                    loftverts[4].x = loftverts[4].y = loftverts[4].z = m_PathVect[pathpos + 1].fScale * (fToScale > 1.0f ? 1.0f : fToScale);
                    fDistRendered += fLength;
                }
                else // split this section in two in order to preserve the taper
                {
                    switch(iCurSegmentSplit)
                    {
                    case 0:
                        loftverts[2] = loftverts[1];
                        loftverts[1] = loftverts[0] + (SNAKE_TAIL_TAPER - fDistRendered) * vDiff / fLength;
                        loftverts[3].x = loftverts[3].y = loftverts[3].z = m_PathVect[pathpos].fScale * fDistRendered / SNAKE_TAIL_TAPER;
                        loftverts[4].x = loftverts[4].y = loftverts[4].z = m_PathVect[pathpos + 1].fScale;
                        ++iCurSegmentSplit;
                        --pathpos;

                        //remove the endto flag if it exists
                        dwFlags &= ~CLofter::LOFT_ENDTO;
                        fLength = SNAKE_TAIL_TAPER - fDistRendered;
                        break;
                    case 1:
                        //loftverts[2] = loftverts[0];
                        loftverts[0] += (SNAKE_TAIL_TAPER - fDistRendered) * vDiff / fLength;
                        loftverts[3].x = loftverts[3].y = loftverts[3].z = m_PathVect[pathpos + 1].fScale;
                        loftverts[4].x = loftverts[4].y = loftverts[4].z = m_PathVect[pathpos + 1].fScale;
                        iCurSegmentSplit = 0;
                        ++iSegDifference;

                        //remove the endfrom flag if it exists
                        dwFlags &= ~CLofter::LOFT_ENDFROM;

                        FLOAT fOldLength = fLength;
                        fLength -= SNAKE_TAIL_TAPER - fDistRendered;
                        fDistRendered += fOldLength;
                        break;
                    };
                }
            }
            else
            {
                fDistRendered += fLength;
                loftverts[3].x = loftverts[3].y = loftverts[3].z = m_PathVect[pathpos].fScale;
                loftverts[4].x = loftverts[4].y = loftverts[4].z = m_PathVect[pathpos + 1].fScale;
            }

            loftverts[3] *= m_fScaleRadius;
            loftverts[4] *= m_fScaleRadius;

            //loft the section of the snake
            pVertices = CLofter::LoftBeveledSectionIndexed(
                                        loftverts, m_pSnakeSect,
                                        iSectSize,
                                        pVertices,
                                        m_Color,
                                        CLofter::LOFT_INDEXED | CLofter::LOFT_SCALEBOTH | dwFlags);
        
            //modify the texture coordinates based on the length of the section
            FLOAT fRepeatOffset = fLength / SNAKE_TEX_REPEAT;
            if((UINT)(pVertices - pVerticesSegBegin) >= iSectSize)
            {
                if(dwFlags & CLofter::LOFT_ENDFROM)
                {
                    for(UINT index = 0; index < iSectSize; ++index)
                    {
                        pVerticesSegBegin->tv = fTexOffset;
                        ++pVerticesSegBegin;
                    }
                }
                
                for(UINT index = 0; index < iSectSize; ++index)
                {
                    if(fTexOffset - fRepeatOffset > 0.0f)
                        pVerticesSegBegin->tv = fTexOffset - fRepeatOffset;
                    else
                        pVerticesSegBegin->tv = 0.0f;
                    ++pVerticesSegBegin;
                }

                fTexOffset -= fRepeatOffset;
            }
            else
            {
                --iSegDifference;
            }
        }

        //copy the indices of the last section to the memory buffer
        //in order to start the next vertex buffer
        memcpy(pLastVertexSect, pVertices - iSectSize, sizeof(CUSTOMVERTEX) * iSectSize);

        m_pVB->Unlock();

        //render the primitive section
        m_pd3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLESTRIP, 0,
                                           pVertices - pVerticesBegin,
                                           0, 2 * iSectSize * (iMaxSegs + iSegDifference) + 1);
        iNumSegsDrawn += iMaxSegs;
    }
    delete[] pLastVertexSect;

    m_fCurLength = fDistRendered;

    //transform the snake head to the correct position and render it
    D3DXVECTOR3 vDiff = *GetLeadPoint() - m_PathVect[m_PathVect.size() - 2].vPos;
    D3DXMATRIX matWorld;
    D3DXMatrixTranslation(&matWorld, GetLeadPoint()->x,
                                     GetLeadPoint()->y,
                                     GetLeadPoint()->z);
    m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );
    
    FLOAT fSnakeAngle;
    if(vDiff.y == 0)
        fSnakeAngle = (vDiff.x > 0) ? -3.1415f / 2.0f : 3.1415f / 2.0f;
    else
        fSnakeAngle = (float)(atan2(-vDiff.x, vDiff.y));

    FLOAT fLenXY = (FLOAT) sqrt(vDiff.x * vDiff.x + vDiff.y * vDiff.y);

    FLOAT fSnakeElev;
    if(fLenXY == 0)
        fSnakeElev = (vDiff.z > 0) ? -3.1415f / 2.0f : 3.1415f / 2.0f;
    else
        fSnakeElev = (float)(-atan2(-vDiff.z, fLenXY));

    //stabilize the snake head elevation to increments of 0.1 radians
    fSnakeElev = (int)((fSnakeElev + 0.05f) / 0.1f) * 0.1f;

    D3DXMatrixRotationZ(&matWorld, fSnakeAngle);
    m_pd3dDevice->MultiplyTransform( D3DTS_WORLD, &matWorld );
    D3DXMatrixRotationX(&matWorld, fSnakeElev);
    m_pd3dDevice->MultiplyTransform( D3DTS_WORLD, &matWorld );

    D3DMATERIAL8 mtrl;
    D3DUtil_InitMaterial(mtrl, ((m_Color >> 16) & 0xff) / 255.0f,
                               ((m_Color >> 8) & 0xff) / 255.0f,
                               (m_Color & 0xff) / 255.0f,
                               ((m_Color >> 24) & 0xff) / 255.0f);
    m_pd3dDevice->SetMaterial(&mtrl);
    m_pSnakeHead->Render( m_pd3dDevice, CD3DMesh::NO_USE_MATERIAL );
    
    D3DXMatrixIdentity(&matWorld);
    m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );
}