#include "snakemanager.h"
#include <crtdbg.h>

CSnakeManager::CSnakeManager(void) : m_pd3dDevice(NULL), m_pPrimarySnake(NULL)
{
    //clear the snake array
    for(int i = 0; i < MAX_SNAKES + 1; ++i)
        m_ppSnakeVect[i] = NULL;
}

CSnakeManager::~CSnakeManager(void)
{
    ClearSnakes();
    m_pd3dDevice = NULL;
}

//adds a snake to the list of snakes
CSnake* CSnakeManager::AddSnake(BOOL bPrimary)
{
    int i = 0;
    while(m_ppSnakeVect[i] && i < MAX_SNAKES)
        ++i;

    //too many snakes in vector
    if(m_ppSnakeVect[i])
        return NULL;

    m_ppSnakeVect[i] = new CSnake();

    if(bPrimary)
        m_pPrimarySnake = m_ppSnakeVect[i];

    //create the device objects if the device is initialized
    if(m_pd3dDevice)
    {
        m_ppSnakeVect[i]->InitDeviceObjects(m_pd3dDevice);
        m_ppSnakeVect[i]->RestoreDeviceObjects();
    }
    return m_ppSnakeVect[i];
}

//gets the primary snake
CSnake* CSnakeManager::GetPrimary()
{
    return m_pPrimarySnake;
}

//gets the number of living snakes left
UINT CSnakeManager::GetNumLivingSnakes()
{
    UINT iNumLiving = 0;
    int i = 0;
    while(m_ppSnakeVect[i] && i < MAX_SNAKES)
    {
        if(!m_ppSnakeVect[i]->GetIsDead())
            ++iNumLiving;
        ++i;
    }
    return iNumLiving;
}

//removes a snake from the list
void CSnakeManager::RemoveSnake(CSnake* pSnake)
{
    //TODO: not yet implemented
    _ASSERT(FALSE);
}

//sets the snakes positions/directions to the given positions/directions
void CSnakeManager::SetSnakePositions(int iCount, D3DXVECTOR3* pSnakeStarts, D3DXVECTOR3* pSnakeDirs)
{
    //initialize the snakes using the given start positions and directions
    int index = 0;
    while(m_ppSnakeVect[index] && index < MAX_SNAKES && index < iCount)
    {
        m_ppSnakeVect[index]->ResetPosition(pSnakeStarts[index], pSnakeDirs[index]);
        ++index;
    }

    //reset any remaining snakes to default positions and directions
    if(m_ppSnakeVect[index] && index < MAX_SNAKES)
    {
        while(m_ppSnakeVect[index] && index < MAX_SNAKES)
        {
            m_ppSnakeVect[index]->ResetPosition(D3DXVECTOR3(0.0f, 0.0f, 1.0f),
                                                D3DXVECTOR3(0.0f, 1.0f, 0.0f));
            ++index;
        }
    }
}

//resets the speed of the snakes
void CSnakeManager::Reset()
{
    //move the snakes and check for collisions
    int i = 0;
    while(m_ppSnakeVect[i] && i < MAX_SNAKES)
    {
        m_ppSnakeVect[i]->ResetState();
        ++i;
    }
}

//moves all of the snakes
void CSnakeManager::MoveSnakes(CD3DMesh* pTerrain)
{
    _ASSERT(pTerrain);

    //move the snakes and check for collisions
    int i = 0;
    while(m_ppSnakeVect[i] && i < MAX_SNAKES)
    {
        m_ppSnakeVect[i]->DoAction(CSnake::ACTION_MOVE, pTerrain);
        ++i;
    }
}

//updates the snakes by checking for collisions
//bAffectPrimaryOnly allows the death to only affect the primary one
// used in networking to avoid non-primary snakes from being killed by lag
//returns whether an item was hit or a snake hit something
BOOL CSnakeManager::CheckCollisions(CCourseLoader* pCourseLoader,
                                    CItemManager* pItemManager,
                                    BOOL bAffectPrimaryOnly)
{
    _ASSERT(pCourseLoader && pItemManager);

    BOOL bHitSomething = FALSE;

    //move the snakes and check for collisions
    int i = 0;
    while(m_ppSnakeVect[i] && i < MAX_SNAKES)
    {
        if(!bAffectPrimaryOnly || m_ppSnakeVect[i] == m_pPrimarySnake)
        {
            //find out whether the snake ate an item
            CItem* pItem = pItemManager->GetItemIntersect(m_ppSnakeVect[i]);
            while(pItem)
            {
                pItem->ModifySnake(m_ppSnakeVect[i]);
                pItem = pItemManager->GetItemIntersect(m_ppSnakeVect[i], pItem);
                bHitSomething = TRUE;
            }
        }

        //check for snake collision or wall collision
        if(FindSnakeCollision(m_ppSnakeVect[i], pCourseLoader))
        {
            if(!bAffectPrimaryOnly || m_ppSnakeVect[i] == m_pPrimarySnake)
                m_ppSnakeVect[i]->DoAction(CSnake::ACTION_DEATH);
            else
                m_ppSnakeVect[i]->DoAction(CSnake::ACTION_STOP);
            bHitSomething = TRUE;
        }

        //do course modifications based on snake positions
        //used for water splashes
        pCourseLoader->UpdateCourse(m_ppSnakeVect[i]);

        ++i;
    }

    return bHitSomething;
}

//finds whether a line segment intersects with a snake and
//returns the endpoints of the snake segment that it intersected with
BOOL CSnakeManager::FindSnakeIntersect(const D3DXVECTOR3& vFrom, const D3DXVECTOR3& vTo,
                                       D3DXVECTOR3* pvSnakeFrom, D3DXVECTOR3* pvSnakeTo)
{
    int i = 0;
    while(m_ppSnakeVect[i])
    {
        if(m_ppSnakeVect[i]->GetCollision(vFrom, vTo, NULL, pvSnakeFrom, pvSnakeTo))
            return TRUE;
        ++i;
    }
    return FALSE;
}

//removes all of the snakes from the vector
void CSnakeManager::ClearSnakes(BOOL bSavePrimary)
{
    int i = 0;
    while(m_ppSnakeVect[i] && i < MAX_SNAKES)
    {
        if(m_ppSnakeVect[i] != m_pPrimarySnake || !bSavePrimary)
            SAFE_DELETE(m_ppSnakeVect[i]);
        ++i;
    }
}

void CSnakeManager::Render()
{
    int i = 0;
    while(m_ppSnakeVect[i] && i < MAX_SNAKES)
    {
        m_ppSnakeVect[i]->Render();
        ++i;
    }
}

HRESULT CSnakeManager::InitDeviceObjects( LPDIRECT3DDEVICE8 pd3dDevice )
{
    _ASSERT(pd3dDevice);

    m_pd3dDevice = pd3dDevice;

    //initialize the snakes
    int i = 0;
    while(m_ppSnakeVect[i] && i < MAX_SNAKES)
    {
        if(FAILED( m_ppSnakeVect[i]->InitDeviceObjects( m_pd3dDevice ) ))
            return E_FAIL;
        ++i;
    }

    return S_OK;
}

HRESULT CSnakeManager::RestoreDeviceObjects()
{
    // Restore the snakes
    int i = 0;
    while(m_ppSnakeVect[i] && i < MAX_SNAKES)
    {
        m_ppSnakeVect[i]->RestoreDeviceObjects();
        ++i;
    }
    return S_OK;
}

HRESULT CSnakeManager::InvalidateDeviceObjects()
{
    int i = 0;
    while(m_ppSnakeVect[i] && i < MAX_SNAKES)
    {
        m_ppSnakeVect[i]->InvalidateDeviceObjects();
        ++i;
    }
    return S_OK;
}

HRESULT CSnakeManager::DeleteDeviceObjects()
{
    //delete the snake
    int i = 0;
    while(m_ppSnakeVect[i] && i < MAX_SNAKES)
    {
        m_ppSnakeVect[i]->DeleteDeviceObjects();
        ++i;
    }
    return S_OK;
}

//finds the collision of the snake and returns whether the
//game is over or not
BOOL CSnakeManager::FindSnakeCollision(CSnake* pSnake, CCourseLoader* pCourseLoader)
{
    _ASSERT(pSnake && pCourseLoader);

    //check for collision with other snakes
    D3DXVECTOR3 vFrom;
    D3DXVECTOR3 vTo;
    pSnake->GetLeadSegment(&vFrom, &vTo);
    vTo += 2.5f * pSnake->GetDirection();

    int i = 0;
    while(m_ppSnakeVect[i] && i < MAX_SNAKES)
    {
        if(m_ppSnakeVect[i]->GetCollision(vFrom, vTo, pSnake))
            return TRUE;
        ++i;
    }

    //check for collision with the wall
    if(pCourseLoader->GetWallIntersect(vFrom, vTo))
        return TRUE;

    return FALSE;
}
