// WallSection.cpp: implementation of the CWallSection class.
//
//////////////////////////////////////////////////////////////////////

#include "WallSection.h"
#include "GameMath.h"
#include "Lofter.h"
#include "dxutil.h"
#include <crtdbg.h>     //for _ASSERT

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CWallSection::CWallSection() : m_pWallVert(NULL),
                               m_iWallVertSize(0),
							   m_fWallHeight(0)
{
}

CWallSection::~CWallSection()
{
    SAFE_DELETE_ARRAY( m_pWallVert );
    m_iWallVertSize = 0;
}

//load the wall section from the file
void CWallSection::LoadFromFile(ifstream& infile)
{
    SAFE_DELETE_ARRAY( m_pWallVert );

    infile >> m_iWallVertSize;
	infile >> m_fWallHeight;
    m_pWallVert= new D3DXVECTOR3[m_iWallVertSize];
    for(UINT iVert = 0; iVert < m_iWallVertSize; ++iVert)
    {
        infile >> m_pWallVert[iVert].x
               >> m_pWallVert[iVert].y
               >> m_pWallVert[iVert].z;
    }
}


//returns whether the line segment intersects a wall or not
BOOL CWallSection::GetWallIntersect(const D3DXVECTOR3 &vFrom, const D3DXVECTOR3 &vTo,
                                    D3DXVECTOR3* pvIntersect, D3DXVECTOR3* pvWallFrom,
                                    D3DXVECTOR3* pvWallTo)
{
    _ASSERT(m_pWallVert && m_iWallVertSize);

	FLOAT fZDelta = 0;

    //check for collision with the wall in the XY plane
    for(UINT i = 0; i < m_iWallVertSize; ++i)
    {
        if(CGameMath::GetLineIntersect(vFrom, vTo, m_pWallVert[i],
                                       m_pWallVert[(i + 1) % m_iWallVertSize],
									   pvIntersect, &fZDelta))
        {
			//check whether collision occurred based on wall height
			if(m_fWallHeight)
			{
				if(m_fWallHeight > 0)
				{
					if(fZDelta > m_fWallHeight || fZDelta < 0.0f)
						continue;
				}
				else
				{
					if(fZDelta < m_fWallHeight || fZDelta > 0.0f)
						continue;
				}
			}
            if(pvWallFrom)
                *pvWallFrom = m_pWallVert[i];
            if(pvWallTo)
                *pvWallTo = m_pWallVert[(i + 1) % m_iWallVertSize];
            return TRUE;
        }
    }
    return FALSE;
}

//returns the distance squared to the nearest wall to the point given
//returns -1.0f if there is no nearest wall
FLOAT CWallSection::GetNearestWall(const D3DXVECTOR3 &vPoint, D3DXVECTOR3* pvWallFrom, D3DXVECTOR3* pvWallTo)
{
    _ASSERT(m_pWallVert && m_iWallVertSize);

    FLOAT fNearDist = -1.0f;
    D3DXVECTOR3 vIntersect;
    for(UINT i = 0; i < m_iWallVertSize; ++i)
    {
        if(CGameMath::GetPointPerp(m_pWallVert[i], m_pWallVert[(i + 1) % m_iWallVertSize], vPoint, &vIntersect))
        {
            vIntersect -= vPoint;
            FLOAT fDist = vIntersect.x * vIntersect.x + vIntersect.y * vIntersect.y;
            if(fNearDist == -1.0f || fDist < fNearDist)
            {
                fNearDist = fDist;
                if(pvWallFrom)
                    *pvWallFrom = m_pWallVert[i];
                if(pvWallTo)
                    *pvWallTo = m_pWallVert[(i + 1) % m_iWallVertSize];
            }
        }
    }
    return fNearDist;
}

//returns a new windows regions defining the wall outline
HRGN CWallSection::CreateWallRegion()
{
    _ASSERT(m_pWallVert && m_iWallVertSize);

    //generate a region in order to place the items in the scene
    POINT* pPoints = new POINT[m_iWallVertSize];
    for(UINT iPoint = 0; iPoint < m_iWallVertSize; ++iPoint)
    {
        pPoints[iPoint].x = (LONG)m_pWallVert[iPoint].x;
        pPoints[iPoint].y = (LONG)m_pWallVert[iPoint].y;
    }
    HRGN hRegion = CreatePolygonRgn(pPoints, m_iWallVertSize, WINDING);
    SAFE_DELETE_ARRAY( pPoints );

    return hRegion;
}

//draw the walls on the given Direct3DDevice for DEBUG purposes only
void CWallSection::Render(LPDIRECT3DDEVICE8 pd3dDevice, const VECTOR2TEX* pWallSect, UINT iWallSectSize)
{
    _ASSERT(pd3dDevice);

    LPDIRECT3DVERTEXBUFFER8 pVBWall;
    //create the lofted object for the wall
    UINT iVbSize     = (2 * iWallSectSize * m_iWallVertSize + 3) * sizeof(CUSTOMVERTEX);
    if( FAILED( pd3dDevice->CreateVertexBuffer( iVbSize, D3DUSAGE_WRITEONLY, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, &pVBWall ) ) )
        return;

    CUSTOMVERTEX* pVertices;
    if( FAILED( pVBWall->Lock( 0, iVbSize, (BYTE**)&pVertices, 0 ) ) )
    {
        SAFE_RELEASE( pVBWall );
        return;
    }

    UINT iVBWallSize = 2 * iWallSectSize * m_iWallVertSize + 1;

    D3DXVECTOR3 loftverts[4];
    DWORD       dwFlags;
    for(UINT i = 0; i < m_iWallVertSize; ++i)
    {
        loftverts[0] = m_pWallVert[i];
        loftverts[1] = m_pWallVert[(i + 1) % m_iWallVertSize];
        loftverts[2] = m_pWallVert[(i + m_iWallVertSize - 1) % m_iWallVertSize];
        loftverts[3] = m_pWallVert[(i + 2) % m_iWallVertSize];
        dwFlags = CLofter::LOFT_DEFAULT;
        if(i == 0)
            dwFlags |= CLofter::LOFT_ENDFROM;
        if(i == m_iWallVertSize - 1)
            dwFlags |= CLofter::LOFT_ENDTO;

        pVertices = CLofter::LoftBeveledSection(loftverts,
                                                pWallSect,
                                                iWallSectSize,
                                                pVertices,
                                                dwFlags);
    }
    pVBWall->Unlock();

    //Render the walls
    pd3dDevice->SetStreamSource( 0, pVBWall, sizeof(CUSTOMVERTEX) );
    pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, iVBWallSize);

    SAFE_RELEASE( pVBWall );
}

/*HRESULT CWallSection::RestoreDeviceObjects(const VECTOR2TEX* pWallSect, UINT iWallSectSize)
{
    _ASSERT(pWallSect);

    if(!m_pd3dDevice || !m_iWallVertSize)
        return S_OK;

    //create the lofted object for the wall
    UINT iVbSize     = (2 * iWallSectSize * m_iWallVertSize + 3) * sizeof(CUSTOMVERTEX);
    if( FAILED( m_pd3dDevice->CreateVertexBuffer( iVbSize, D3DUSAGE_WRITEONLY, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, &m_pVBWall ) ) )
        return E_FAIL;

    CUSTOMVERTEX* pVertices;
    if( FAILED( m_pVBWall->Lock( 0, iVbSize, (BYTE**)&pVertices, 0 ) ) )
    {
        m_iVBWallSize = 0;
        return E_FAIL;
    }

    m_iVBWallSize = 2 * iWallSectSize * m_iWallVertSize + 1;

    D3DXVECTOR3 loftverts[4];
    DWORD       dwFlags;
    for(UINT i = 0; i < m_iWallVertSize; ++i)
    {
        loftverts[0] = m_pWallVert[i];
        loftverts[1] = m_pWallVert[(i + 1) % m_iWallVertSize];
        loftverts[2] = m_pWallVert[(i + m_iWallVertSize - 1) % m_iWallVertSize];
        loftverts[3] = m_pWallVert[(i + 2) % m_iWallVertSize];
        dwFlags = CLofter::LOFT_DEFAULT;
        if(i == 0)
            dwFlags |= CLofter::LOFT_ENDFROM;
        if(i == m_iWallVertSize - 1)
            dwFlags |= CLofter::LOFT_ENDTO;

        pVertices = CLofter::LoftBeveledSection(loftverts,
                                                pWallSect,
                                                iWallSectSize,
                                                pVertices,
                                                dwFlags);
    }
    m_pVBWall->Unlock();

    return S_OK;
}

HRESULT CWallSection::InvalidateDeviceObjects()
{
    SAFE_RELEASE( m_pVBWall );
    return S_OK;
}

HRESULT CWallSection::DeleteDeviceObjects()
{
    SAFE_RELEASE( m_pVBWall );
    SAFE_DELETE_ARRAY( m_pWallVert );
    m_iWallVertSize = 0;
    return S_OK;
}*/
