// WaterPlane.cpp: implementation of the CWaterPlane class.
//
//////////////////////////////////////////////////////////////////////

#include "WaterPlane.h"
#include "crtdbg.h"     //for _ASSERT
#include "DXUtil.h"

// Custom D3D vertex format used by the vertex buffer
struct WATERPLANEVERTEX
{
    D3DXVECTOR3 position;       // vertex position
    D3DCOLOR    color;          // vertex color
    FLOAT       tu, tv;         // The texture coordinates.
};

#define D3DFVF_WATERPLANEVERTEX (D3DFVF_XYZ|D3DFVF_DIFFUSE|D3DFVF_TEX1)

#define PI 3.1415

//table of sins and cos of different angles in 1/8 increments
const D3DXVECTOR2 g_vAngleOffsets[] =
{
    D3DXVECTOR2((float)cos(0.0),        (float)sin(0.0)),
    D3DXVECTOR2((float)cos(PI / 16),    (float)sin(PI / 16)),
    D3DXVECTOR2((float)cos(PI / 8),     (float)sin(PI / 8)),
    D3DXVECTOR2((float)cos(3 * PI / 16),(float)sin(3 * PI / 16)),
    D3DXVECTOR2((float)cos(PI / 4),     (float)sin(PI / 4)),
    D3DXVECTOR2((float)cos(5 * PI / 16),(float)sin(5 * PI / 16)),
    D3DXVECTOR2((float)cos(3 * PI / 8), (float)sin(3 * PI / 8)),
    D3DXVECTOR2((float)cos(7 * PI / 16),(float)sin(7 * PI / 16)),
    D3DXVECTOR2((float)cos(PI / 2),     (float)sin(PI / 2)),
    D3DXVECTOR2((float)cos(9 * PI / 16),(float)sin(9 * PI / 16)),
    D3DXVECTOR2((float)cos(5 * PI / 8), (float)sin(5 * PI / 8)),
    D3DXVECTOR2((float)cos(11 * PI / 16),(float)sin(11 * PI / 16)),
    D3DXVECTOR2((float)cos(3 * PI / 4), (float)sin(3 * PI / 4)),
    D3DXVECTOR2((float)cos(7 * PI / 8), (float)sin(7 * PI / 8)),
    D3DXVECTOR2((float)cos(PI),         (float)sin(PI)),
    D3DXVECTOR2((float)cos(9 * PI / 8), (float)sin(9 * PI / 8)),
    D3DXVECTOR2((float)cos(5 * PI / 4), (float)sin(5 * PI / 4)),
    D3DXVECTOR2((float)cos(11 * PI / 8), (float)sin(12 * PI / 8)),
    D3DXVECTOR2((float)cos(3 * PI / 2), (float)sin(3 * PI / 2)),
    D3DXVECTOR2((float)cos(13 * PI / 8), (float)sin(13 * PI / 8)),
    D3DXVECTOR2((float)cos(7 * PI / 4), (float)sin(7 * PI / 4)),
    D3DXVECTOR2((float)cos(15 * PI / 8), (float)sin(15 * PI / 8)),
};

const UINT iAngleOffsetSize = sizeof(g_vAngleOffsets) / sizeof(D3DXVECTOR2);

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CWaterPlane::CWaterPlane(const D3DXVECTOR3& vPos,
                         const D3DXVECTOR2&  vSize,
                         const UINT iHorzDiv, const UINT iVertDiv) :
    m_pd3dDevice(NULL), m_pWaterVB(NULL), m_pWaterIB(NULL),
    m_pWaterTexture(NULL), m_pSplashVB(NULL), m_pSplashTexture(NULL),
    m_fWidth(vSize.x), m_fHeight(vSize.y),
    m_iHorzDiv(iHorzDiv), m_iVertDiv(iVertDiv)
{
    //set the pos to the place where upper left corner is
    m_vPos.x = vPos.x - m_fWidth / 2.0f;
    m_vPos.y = vPos.y - m_fHeight / 2.0f;
    m_vPos.z = vPos.z;

    m_pOffsetDirs    = new BOOL[(m_iHorzDiv + 1)*(m_iVertDiv + 1)];
    m_pOffsetIndices = new WORD[(m_iHorzDiv + 1)*(m_iVertDiv + 1)];
    for(UINT i = 0; i < (m_iHorzDiv + 1)*(m_iVertDiv + 1); ++i)
    {
        m_pOffsetIndices[i] = (WORD)((float)(rand() * iAngleOffsetSize) / RAND_MAX);
        m_pOffsetDirs[i]    = rand() > RAND_MAX / 2;
    }
}

CWaterPlane::~CWaterPlane()
{
    delete[] m_pOffsetIndices;
    delete[] m_pOffsetDirs;
}

//updates the splash list for the water
void CWaterPlane::UpdateSplashes(CSnake* pSnake)
{
    D3DXVECTOR3 vFrom;
    D3DXVECTOR3 vTo;
    pSnake->GetLeadSegment(&vFrom, &vTo);
    
    //decrement the splash stage and remove finished splashes
    vector<SplashInfo>::iterator it;
    for(it = m_SplashVect.begin(); it != m_SplashVect.end(); ++it)
    {
        --it->m_iStage;
        if(it->m_iStage == 0)
        {
            it = m_SplashVect.erase(it);
            --it;
        }
    }

    //calculate the intersection with the z-aligned water plane
    FLOAT fVal = (m_vPos.z - vFrom.z) / (vTo.z - vFrom.z);
    if(fVal <= 1 && fVal >= 0)
    {
        //check whether it is in the bounds of the water plane
        SplashInfo splash;
        splash.m_vPos.x = vFrom.x + fVal * (vTo.x - vFrom.x);
        splash.m_vPos.y = vFrom.y + fVal * (vTo.y - vFrom.y);

        //if the position is outside of the plane in the horizontal direction
        //don't add it to the splash list
        if(splash.m_vPos.x < m_vPos.x || splash.m_vPos.x > m_vPos.x + m_fWidth)
            return;

        //if the position is outside of the plane in the vertical direction
        //don't add it to the splash list
        if(splash.m_vPos.y < m_vPos.y || splash.m_vPos.y > m_vPos.y + m_fHeight)
            return;

        splash.m_iStage = 100;   //persist for 100 frames

        D3DXVECTOR2 vDiff;
        for(it = m_SplashVect.begin(); it != m_SplashVect.end(); ++it)
        {
            D3DXVec2Subtract(&vDiff, &it->m_vPos, &splash.m_vPos);
            if(D3DXVec2LengthSq(&vDiff) < 1.0f)
            {
                //reset the persistence time for the already existing splash
                it->m_iStage = 100;
                return;
            }
        }
        m_SplashVect.push_back(splash);
    }
}

HRESULT CWaterPlane::InitDeviceObjects(LPDIRECT3DDEVICE8 pd3dDevice)
{
    _ASSERT(pd3dDevice);

    m_pd3dDevice = pd3dDevice;
    return S_OK;
}

HRESULT CWaterPlane::RestoreDeviceObjects()
{
    _ASSERT(m_pd3dDevice);

    HRESULT hr;

    // Create the vertex buffer for the water
    if( FAILED( hr = m_pd3dDevice->CreateVertexBuffer(
                                                  (m_iHorzDiv + 1)*(m_iVertDiv + 1)*sizeof(WATERPLANEVERTEX),
                                                  D3DUSAGE_WRITEONLY | D3DUSAGE_DYNAMIC,
                                                  D3DFVF_WATERPLANEVERTEX,
                                                  D3DPOOL_DEFAULT, &m_pWaterVB ) ) )
        return E_FAIL;

    // Fill the vertex buffer with triangles
    WATERPLANEVERTEX* pVertices;

    if( FAILED( hr = m_pWaterVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) )
        return E_FAIL;

    FLOAT fHorzInc = m_fWidth / m_iHorzDiv;
    FLOAT fVertInc = m_fHeight / m_iVertDiv;
    for(UINT ix = 0; ix <= m_iHorzDiv; ++ix)
    {
        for(UINT iy = 0; iy <= m_iVertDiv; ++iy)
        {
            pVertices->position.x = m_vPos.x + ix * fHorzInc;
            pVertices->position.y = m_vPos.y + iy * fVertInc;
            pVertices->position.z = m_vPos.z;
            pVertices->color = D3DCOLOR_RGBA(255,255,255,255);
            pVertices->tu = (float)ix / m_iHorzDiv;
            pVertices->tv = (float)iy / m_iVertDiv;
            ++pVertices;
        }
    }

    m_pWaterVB->Unlock();

    //create the index buffer for the water
    UINT iIndexBufSize = 2 * (m_iVertDiv + 1) * m_iHorzDiv - (m_iHorzDiv - 1);
    if( FAILED( m_pd3dDevice->CreateIndexBuffer( iIndexBufSize * sizeof(WORD),
                                                 D3DUSAGE_WRITEONLY, D3DFMT_INDEX16,
                                                 D3DPOOL_DEFAULT, &m_pWaterIB ) ) )
        return E_FAIL;

    WORD* pIndices;
    //lock the index buffer for the water
    if(FAILED( m_pWaterIB->Lock(0, 0, (BYTE**)&pIndices, 0 ) ))
        return E_FAIL;

    //generate all connected columns into one triangle strip
    for(UINT iCol = 0; iCol < m_iHorzDiv; ++iCol)
    {
        //generate one column of the plane
        if(iCol % 2)
        {
            UINT iStart = iCol * (m_iVertDiv + 1) + m_iVertDiv;
            for(UINT iRow = 0; iRow < m_iVertDiv; ++iRow)
            {
                *pIndices = iStart - iRow;
                ++pIndices;
                *pIndices = iStart - iRow + m_iVertDiv + 1;
                ++pIndices;
            }
            *pIndices = iCol * (m_iVertDiv + 1);
        }
        else
        {
            UINT iStart = iCol * (m_iVertDiv + 1);
            for(UINT iRow = 0; iRow < m_iVertDiv; ++iRow)
            {
                *pIndices = iStart + iRow;
                ++pIndices;
                *pIndices = iStart + iRow + m_iVertDiv + 1;
                ++pIndices;
            }
            *pIndices = iCol * (m_iVertDiv + 1) + m_iVertDiv;
        }
        ++pIndices;
    }
    if(m_iHorzDiv % 2)
    {
        *pIndices = (m_iHorzDiv + 1)*(m_iVertDiv + 1) - 1;
    }
    else
    {
        *pIndices = m_iHorzDiv * (m_iVertDiv + 1);
    }
    ++pIndices;

    m_pWaterIB->Unlock();

    //create the texture for the water from the file
    if(FAILED( D3DXCreateTextureFromFile(m_pd3dDevice, "Core\\water.png", &m_pWaterTexture) ))
        return E_FAIL;

    // Create the vertex buffer for the splash
    if( FAILED( hr = m_pd3dDevice->CreateVertexBuffer(4 * sizeof(WATERPLANEVERTEX),
                                                  D3DUSAGE_WRITEONLY,
                                                  D3DFVF_WATERPLANEVERTEX,
                                                  D3DPOOL_DEFAULT, &m_pSplashVB ) ) )
        return E_FAIL;

    // Fill the vertex buffer to create a flat square for the splash
    if( FAILED( hr = m_pSplashVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) )
        return E_FAIL;

    pVertices->position = D3DXVECTOR3(5.0f, 5.0f, 0.0f);
    pVertices->color = D3DCOLOR_RGBA(255,255,255,255);
    pVertices->tu = 0.0f;
    pVertices->tv = 1.0f;
    ++pVertices;
    pVertices->position = D3DXVECTOR3(-5.0f, 5.0f, 0.0f);
    pVertices->color = D3DCOLOR_RGBA(255,255,255,255);
    pVertices->tu = 1.0f;
    pVertices->tv = 1.0f;
    ++pVertices;
    pVertices->position = D3DXVECTOR3(5.0f, -5.0f, 0.0f);
    pVertices->color = D3DCOLOR_RGBA(255,255,255,255);
    pVertices->tu = 0.0f;
    pVertices->tv = 0.0f;
    ++pVertices;
    pVertices->position = D3DXVECTOR3(-5.0f, -5.0f, 0.0f);
    pVertices->color = D3DCOLOR_RGBA(255,255,255,255);
    pVertices->tu = 1.0f;
    pVertices->tv = 0.0f;
    ++pVertices;

    m_pSplashVB->Unlock();

    //create the texture for the water from the file
    if(FAILED( D3DXCreateTextureFromFile(m_pd3dDevice, "Core\\waterring.png", &m_pSplashTexture) ))
        return E_FAIL;

    return S_OK;
}

HRESULT CWaterPlane::InvalidateDeviceObjects()
{
    m_pd3dDevice->SetTexture( 0, NULL );
    SAFE_RELEASE( m_pWaterVB );
    SAFE_RELEASE( m_pWaterIB );
    SAFE_RELEASE( m_pWaterTexture );
    SAFE_RELEASE( m_pSplashVB );
    SAFE_RELEASE( m_pSplashTexture );
    return S_OK;
}

HRESULT CWaterPlane::DeleteDeviceObjects()
{
    m_pd3dDevice->SetTexture( 0, NULL );
    SAFE_RELEASE( m_pWaterVB );
    SAFE_RELEASE( m_pWaterIB );
    SAFE_RELEASE( m_pWaterTexture );
    SAFE_RELEASE( m_pSplashVB );
    SAFE_RELEASE( m_pSplashTexture );
    m_pd3dDevice = NULL;
    return S_OK;
}

void CWaterPlane::Render()
{
    _ASSERT(m_pd3dDevice);

    m_pd3dDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
    m_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
    m_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );

    /////////////////////////////////////////////////////
    // draw the water
    /////////////////////////////////////////////////////

    // Render the vertex buffer contents
    m_pd3dDevice->SetStreamSource( 0, m_pWaterVB, sizeof(WATERPLANEVERTEX) );
    m_pd3dDevice->SetVertexShader( D3DFVF_WATERPLANEVERTEX );
    m_pd3dDevice->SetIndices( m_pWaterIB, 0 );
    m_pd3dDevice->SetTexture( 0, m_pWaterTexture );

    m_pd3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
    m_pd3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1 );

    m_pd3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLESTRIP, 0,
                                       (m_iHorzDiv + 1)*(m_iVertDiv + 1),
                                       0, 2 * m_iVertDiv * m_iHorzDiv + m_iHorzDiv - 1);

    /////////////////////////////////////////////////////
    // draw the splashes
    /////////////////////////////////////////////////////
    m_pd3dDevice->SetStreamSource( 0, m_pSplashVB, sizeof(WATERPLANEVERTEX) );
    m_pd3dDevice->SetTexture( 0, m_pSplashTexture );

    D3DXMATRIX matWorld;
    D3DXMATRIX matScale;
    vector<SplashInfo>::iterator it;
    for(it = m_SplashVect.begin(); it != m_SplashVect.end(); ++it)
    {
        D3DXMatrixScaling(&matScale, 1.0f - it->m_iStage / 100.0f, 1.0f - it->m_iStage / 100.0f, 1.0f - it->m_iStage / 100.0f);
        D3DXMatrixTranslation(&matWorld, it->m_vPos.x, it->m_vPos.y, m_vPos.z);
        D3DXMatrixMultiply(&matWorld, &matScale, &matWorld); 
        m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );
        m_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
    }
    
    //reset the state
    m_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE, TRUE );
    m_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
    m_pd3dDevice->SetRenderState(D3DRS_LIGHTING, TRUE);
}

HRESULT CWaterPlane::FrameMove()
{
    //increment/decrement the index into the angle offset
    //array based on the direction of rotation
    WORD* pCurOffset = m_pOffsetIndices;
    BOOL* pCurDir    = m_pOffsetDirs;
    for(UINT i = 0; i < (m_iHorzDiv + 1)*(m_iVertDiv + 1); ++i)
    {
        if(*pCurDir)
        {
            ++(*pCurOffset);
            *pCurOffset %= iAngleOffsetSize;
        }
        else
        {
            if(*pCurOffset)
                --(*pCurOffset);
            else
                *pCurOffset = iAngleOffsetSize - 1;
        }
        ++pCurOffset;
        ++pCurDir;
    }

    // Change the vertex buffer based on the angle offsets
    WATERPLANEVERTEX* pVertices;
    HRESULT hr;
    if( FAILED( hr = m_pWaterVB->Lock( 0, 0, (BYTE**)&pVertices, 0 ) ) )
        return E_FAIL;

    FLOAT fHorzInc = m_fWidth / m_iHorzDiv;
    FLOAT fVertInc = m_fHeight / m_iVertDiv;

    //skip the first and last columns so edges don't get purturbed
    pVertices += m_iVertDiv + 1;
    for(UINT ix = 1; ix < m_iHorzDiv; ++ix)
    {
        //skip the first item of every column so edges don't get purturbed
        ++pVertices;

        for(UINT iy = 1; iy < m_iVertDiv; ++iy)
        {
            pVertices->position.x = m_vPos.x + ix * fHorzInc +
                                    0.2f * fHorzInc * g_vAngleOffsets[m_pOffsetIndices[ix*(m_iVertDiv + 1) + iy]].x;
            pVertices->position.y = m_vPos.y + iy * fVertInc +
                                    0.2f * fVertInc * g_vAngleOffsets[m_pOffsetIndices[ix*(m_iVertDiv + 1) + iy]].y;
            pVertices->position.z = m_vPos.z; //+ 0.1f * fVertInc * g_vAngleOffsets[m_pOffsetIndices[ix*(m_iVertDiv + 1) + iy]].y;
            ++pVertices;
        }

        //skip the last item of every column
        ++pVertices;
    }
    m_pWaterVB->Unlock();

    return S_OK;
}